from zrb.helper.accessories.color import colored
from zrb.helper.log import logger
from zrb.helper.string.untyped_conversion import (
    untyped_to_boolean,
    untyped_to_cli_name,
    untyped_to_logging_level,
    untyped_to_variable_name,
)
from zrb.helper.typecheck import typechecked

logger.debug(colored("Loading zrb.helper.string.conversion", attrs=["dark"]))


@typechecked
def to_cli_name(name: str) -> str:
    return untyped_to_cli_name((name))


@typechecked
def to_variable_name(string: str) -> str:
    return untyped_to_variable_name(string)


@typechecked
def to_boolean(string: str) -> bool:
    return untyped_to_boolean(string)


@typechecked
def to_logging_level(logging_level_str: str) -> int:
    return untyped_to_logging_level(logging_level_str)



from zrb import PathChecker

from .._constant import APP_FRONTEND_BUILD_DIR
from ._group import snake_zrb_app_name_frontend_group

check_frontend_path = PathChecker(
    name="check-frontend-path",
    group=snake_zrb_app_name_frontend_group,
    path=APP_FRONTEND_BUILD_DIR,
)



import os

import aiofiles

from zrb.helper.accessories.color import colored
from zrb.helper.log import logger
from zrb.helper.typecheck import typechecked

logger.debug(colored("Loading zrb.helper.file.text", attrs=["dark"]))


@typechecked
async def read_text_file_async(file_name: str) -> str:
    async with aiofiles.open(file_name, mode="r") as file:
        content = await file.read()
    return content


@typechecked
async def write_text_file_async(file_name: str, content: str):
    os.makedirs(os.path.dirname(file_name), exist_ok=True)
    async with aiofiles.open(file_name, mode="w") as file:
        await file.write(content)


@typechecked
async def append_text_file_async(file_name: str, additional_content: str):
    content = await read_text_file_async(file_name)
    if content != "":
        additional_content = "\n" + additional_content
    new_content = content + additional_content
    await write_text_file_async(file_name, new_content)



import os

from zrb.helper.file.text import read_text_file_async, write_text_file_async
from zrb.helper.typecheck import typechecked
from zrb.helper.util import to_capitalized_human_readable, to_kebab_case, to_snake_case
from zrb.task.task import Task

from ._common import (
    get_app_frontend_routes_dir,
    get_default_js_value,
    get_html_input,
    replace_marker,
)


@typechecked
async def inject_insert_page(
    task: Task,
    project_dir: str,
    app_name: str,
    module_name: str,
    entity_name: str,
    column_name: str,
    column_type: str,
):
    kebab_module_name = to_kebab_case(module_name)
    kebab_entity_name = to_kebab_case(entity_name)
    snake_column_name = to_snake_case(column_name)
    kebab_column_name = to_kebab_case(column_name)
    column_caption = to_capitalized_human_readable(column_name)
    file_path = os.path.join(
        get_app_frontend_routes_dir(project_dir, app_name),
        kebab_module_name,
        kebab_entity_name,
        "new",
        "+page.svelte",
    )
    task.print_out(f"Read HTML from: {file_path}")
    html_content = await read_text_file_async(file_path)
    task.print_out("Add default value to insert page")
    default_value = get_default_js_value(column_type)
    html_content = replace_marker(
        html_content,
        marker="// DON'T DELETE: set field default value here",
        code=f"row.{snake_column_name} = {default_value}",
    )
    task.print_out("Add field to insert page")
    html_input = get_html_input(column_type, column_name, column_caption)
    html_content = replace_marker(
        html_content,
        marker="<!-- DON'T DELETE: insert new field here-->",
        code="\n".join(
            [
                '<div class="mb-4">',
                f'    <label class="block text-gray-700 font-bold mb-2" for="{kebab_column_name}">{column_caption}</label>',  # noqa
                f"    {html_input}",
                "</div>",
            ]
        ),
    )
    task.print_out(f"Write modified HTML to: {file_path}")
    await write_text_file_async(file_path, html_content)



from collections.abc import Mapping
from typing import Any

from zrb.helper.accessories.color import colored
from zrb.helper.log import logger
from zrb.helper.typecheck import typechecked

logger.debug(
    colored("Loading zrb.helper.docker_compose.fetch_external_env", attrs=["dark"])
)


@typechecked
def fetch_compose_file_env_map(data: Any) -> Mapping[str, str]:
    global_env_dict = {}
    if "services" not in data:
        return global_env_dict
    for service in data["services"]:
        if "environment" not in data["services"][service]:
            continue
        environments = data["services"][service]["environment"]
        if isinstance(environments, list):
            for environment in environments:
                parts: list[str] = environment.split("=")
                if len(parts) > 0:
                    env_str = str(parts[1])
                    env_dict = parse_compose_file_env_string(env_str)
                    global_env_dict.update(env_dict)
        if isinstance(environments, dict):
            for _, env_str in environments.items():
                env_str = str(env_str)
                env_dict = parse_compose_file_env_string(env_str)
                global_env_dict.update(env_dict)
    return global_env_dict


@typechecked
def parse_compose_file_env_string(env_str: str) -> Mapping[str, str]:
    env_dict = {}
    stack = []
    key = ""
    value = ""
    index = 0
    while index < len(env_str):
        char = env_str[index]
        if char == "{":
            stack.append(index)
        elif char == "}":
            start = stack.pop()
            if not stack:
                segment = env_str[start + 1 : index]
                if ":-" in segment:
                    key, value = segment.split(":-", 1)
                    if value.startswith("${") and value.endswith("}"):
                        sub_dict = parse_compose_file_env_string(value)
                        env_dict.update(sub_dict)
                else:
                    key = segment
                    value = ""
                if "${" in value:
                    value = ""
                env_dict[key] = value
                key = ""
                value = ""
        index += 1
    return env_dict



from zrb.builtin.monorepo.add import add_to_monorepo
from zrb.builtin.monorepo.pull import pull_from_monorepo
from zrb.builtin.monorepo.push import push_to_monorepo

assert add_to_monorepo
assert pull_from_monorepo
assert push_to_monorepo



from zrb.builtin.project.add._group import project_add_group
from zrb.task_group.group import Group

project_add_app_group = Group(
    name="app", description="Add app/generator to project", parent=project_add_group
)



import os

from zrb.builtin.devtool.install._group import devtool_install_group
from zrb.builtin.devtool.install._helper import write_config
from zrb.builtin.devtool.install._input import terminal_config_file_input
from zrb.runner import runner
from zrb.task.cmd_task import CmdTask
from zrb.task.flow_task import FlowTask
from zrb.task.task import Task

_CURRENT_DIR = os.path.dirname(__file__)
_SHELL_SCRIPT_DIR = os.path.join(_CURRENT_DIR, "..", "..", "..", "..", "shell-scripts")

install_pulumi = FlowTask(
    name="pulumi",
    group=devtool_install_group,
    description="Universal infrastructure as code",
    inputs=[terminal_config_file_input],
    steps=[
        CmdTask(
            name="install-pulumi",
            cmd_path=[
                os.path.join(_SHELL_SCRIPT_DIR, "_common-util.sh"),
                os.path.join(_CURRENT_DIR, "install.sh"),
            ],
            preexec_fn=None,
        ),
        Task(
            name="configure-pulumi",
            run=write_config(
                template_file=os.path.join(_CURRENT_DIR, "resource", "config.sh"),
                config_file="{{input.config_file}}",
            ),
        ),
    ],
    retry=0,
)
runner.register(install_pulumi)



from typing import Any

from zrb.config.config import VERSION
from zrb.runner import runner
from zrb.task.decorator import python_task
from zrb.task.task import Task


@python_task(name="version", description="Get Zrb version", runner=runner)
async def get_version(*args: Any, **kwargs: Any) -> str:
    task: Task = kwargs.get("_task")
    task.print_out(VERSION)
    return VERSION



from component.repo import DBEntityMixin, DBRepo, Repo
from module.log.integration import Base
from module.log.schema.activity import Activity, ActivityData
from sqlalchemy import Column, String, Text


class DBEntityActivity(Base, DBEntityMixin):
    class Config:
        from_attributes = True

    __tablename__ = "activities"
    action = Column(String)
    entity: Column = Column(String)
    data: Column = Column(Text)


class ActivityRepo(Repo[Activity, ActivityData]):
    pass


class ActivityDBRepo(DBRepo[Activity, ActivityData], ActivityRepo):
    schema_cls = Activity
    db_entity_cls = DBEntityActivity



from zrb import DockerComposeTask, runner

from ..._project import publish_project, push_project_images
from .._constant import RESOURCE_DIR
from .._input import local_input
from ._env import image_env
from ._group import snake_zrb_app_name_image_group
from ._input import image_input
from .build import build_snake_zrb_app_name_image

push_snake_zrb_app_name_image = DockerComposeTask(
    icon="📰",
    name="push",
    description="Push human readable zrb app name image",
    group=snake_zrb_app_name_image_group,
    inputs=[
        local_input,
        image_input,
    ],
    envs=[image_env],
    upstreams=[build_snake_zrb_app_name_image],
    cwd=RESOURCE_DIR,
    compose_cmd="push",
    compose_env_prefix="CONTAINER_ZRB_ENV_PREFIX",
)

push_snake_zrb_app_name_image >> push_project_images
push_snake_zrb_app_name_image >> publish_project

runner.register(push_snake_zrb_app_name_image)



from config import (
    APP_AUTH_JWT_TOKEN_ALGORITHM,
    APP_AUTH_JWT_TOKEN_SECRET_KEY,
    APP_AUTH_REFRESH_TOKEN_TYPE,
)
from module.auth.component import JWTRefreshTokenUtil, RefreshTokenUtil


def init_token_util() -> RefreshTokenUtil:
    if APP_AUTH_REFRESH_TOKEN_TYPE.lower() == "jwt":
        return JWTRefreshTokenUtil(
            secret_key=APP_AUTH_JWT_TOKEN_SECRET_KEY,
            algorithm=APP_AUTH_JWT_TOKEN_ALGORITHM,
        )
    raise ValueError(f"Invalid auth token type: {APP_AUTH_REFRESH_TOKEN_TYPE}")


refresh_token_util = init_token_util()



from component.rpc import Caller
from module.auth.component.authorizer.authorizer import Authorizer


class RPCAuthorizer(Authorizer):
    def __init__(
        self,
        rpc_caller: Caller,
        is_admin_rpc_name: str,
        is_guest_rpc_name: str,
        is_user_authorized_rpc_name: str,
    ):
        self.rpc_caller = rpc_caller
        self.is_admin_rpc_name = is_admin_rpc_name
        self.is_guest_rpc_name = is_guest_rpc_name
        self.is_user_authorized_rpc_name = is_user_authorized_rpc_name

    async def is_admin(self, user_id: str) -> bool:
        return await self.rpc_caller.call(self.is_admin_rpc_name, user_id)

    async def is_guest(self, user_id: str) -> bool:
        return await self.rpc_caller.call(self.is_guest_rpc_name, user_id)

    async def is_having_permission(self, user_id: str, *permission_name: str) -> bool:
        permission_map = await self.rpc_caller.call(
            self.is_user_authorized_rpc_name,
            id=user_id,
            permission_name=permission_name,
        )
        for permission in permission_map:
            if not permission_map[permission]:
                return False
        return True



import asyncio
import datetime
from collections.abc import Callable, Iterable, Mapping
from typing import Any, Optional, TypeVar, Union

import croniter

from zrb.helper.accessories.color import colored
from zrb.helper.log import logger
from zrb.helper.typecheck import typechecked
from zrb.helper.typing import JinjaTemplate
from zrb.task.any_task import AnyTask
from zrb.task.any_task_event_handler import (
    OnFailed,
    OnReady,
    OnRetry,
    OnSkipped,
    OnStarted,
    OnTriggered,
    OnWaiting,
)
from zrb.task.watcher import Watcher
from zrb.task_env.env import Env
from zrb.task_env.env_file import EnvFile
from zrb.task_group.group import Group
from zrb.task_input.any_input import AnyInput

logger.debug(colored("Loading zrb.task.time_watcher", attrs=["dark"]))

TTimeWatcher = TypeVar("TTimeWatcher", bound="TimeWatcher")


@typechecked
class TimeWatcher(Watcher):
    """
    TimeWatcher will wait for any changes specified on  path.

    Once the changes detected, TimeWatcher will be completed
    and <task-name>.scheduled-time xcom will be set.
    """

    __scheduled_times: Mapping[str, Mapping[str, datetime.datetime]] = {}

    def __init__(
        self,
        name: str = "watch-path",
        group: Optional[Group] = None,
        inputs: Iterable[AnyInput] = [],
        envs: Iterable[Env] = [],
        env_files: Iterable[EnvFile] = [],
        icon: Optional[str] = None,
        color: Optional[str] = None,
        description: str = "",
        upstreams: Iterable[AnyTask] = [],
        fallbacks: Iterable[AnyTask] = [],
        on_triggered: Optional[OnTriggered] = None,
        on_waiting: Optional[OnWaiting] = None,
        on_skipped: Optional[OnSkipped] = None,
        on_started: Optional[OnStarted] = None,
        on_ready: Optional[OnReady] = None,
        on_retry: Optional[OnRetry] = None,
        on_failed: Optional[OnFailed] = None,
        schedule: JinjaTemplate = "",
        checking_interval: Union[int, float] = 0.05,
        progress_interval: Union[int, float] = 30,
        should_execute: Union[bool, JinjaTemplate, Callable[..., bool]] = True,
    ):
        Watcher.__init__(
            self,
            name=name,
            group=group,
            inputs=inputs,
            envs=envs,
            env_files=env_files,
            icon=icon,
            color=color,
            description=description,
            upstreams=upstreams,
            fallbacks=fallbacks,
            on_triggered=on_triggered,
            on_waiting=on_waiting,
            on_skipped=on_skipped,
            on_started=on_started,
            on_ready=on_ready,
            on_retry=on_retry,
            on_failed=on_failed,
            checking_interval=checking_interval,
            progress_interval=progress_interval,
            should_execute=should_execute,
        )
        self._schedule = schedule
        self._rendered_schedule: str = ""

    def copy(self) -> TTimeWatcher:
        return super().copy()

    def to_function(
        self,
        env_prefix: str = "",
        raise_error: bool = True,
        is_async: bool = False,
        show_done_info: bool = True,
        should_clear_xcom: bool = False,
        should_stop_looper: bool = False,
    ) -> Callable[..., bool]:
        return super().to_function(
            env_prefix=env_prefix,
            raise_error=raise_error,
            is_async=is_async,
            show_done_info=show_done_info,
            should_clear_xcom=should_clear_xcom,
            should_stop_looper=should_stop_looper,
        )

    async def run(self, *args: Any, **kwargs: Any) -> bool:
        self._rendered_schedule = self.render_str(self._schedule)
        identifier = self.get_identifier()
        if identifier not in self.__scheduled_times:
            self.__scheduled_times[identifier] = self._get_next_schedule_time()
        return await super().run(*args, **kwargs)

    def create_loop_inspector(self) -> Callable[..., Optional[bool]]:
        async def loop_inspect() -> bool:
            await asyncio.sleep(self._checking_interval)
            label = f"Watching {self._rendered_schedule}"
            identifier = self.get_identifier()
            scheduled_time = self.__scheduled_times[identifier]
            self.set_task_xcom(key="scheduled-time", value=scheduled_time)
            now = datetime.datetime.now()
            if now > scheduled_time:
                self.print_out_dark(f"{label} (Meet {scheduled_time})")
                self.__scheduled_times[identifier] = self._get_next_schedule_time()
                return True
            self.show_progress(f"{label} (Waiting for {scheduled_time})")
            return False

        return loop_inspect

    def _get_next_schedule_time(self) -> datetime.datetime:
        cron = self._get_cron()
        return cron.get_next(datetime.datetime)

    def _get_cron(self) -> Any:
        margin = datetime.timedelta(seconds=self._checking_interval / 2.0)
        slightly_before_now = datetime.datetime.now() - margin
        cron = croniter.croniter(self._rendered_schedule, slightly_before_now)
        return cron



from typing import Optional

import libcst as cst

from zrb.helper.accessories.color import colored
from zrb.helper.log import logger
from zrb.helper.typecheck import typechecked

logger.debug(
    colored("Loading zrb.helper.codemod.add_property_to_class", attrs=["dark"])
)


@typechecked
class AddPropertyTransformer(cst.CSTTransformer):
    def __init__(
        self,
        class_name: str,
        property_name_node: cst.Name,
        property_type_node: cst.Annotation,
        property_value_node: Optional[cst.BaseExpression],
    ):
        self.class_name = class_name
        self.property_name_node = property_name_node
        self.property_type_node = property_type_node
        self.property_value_node = property_value_node
        super().__init__()

    def leave_ClassDef(
        self, original_node: cst.ClassDef, updated_node: cst.ClassDef
    ) -> cst.ClassDef:
        if updated_node.name.value != self.class_name:
            return updated_node
        class_body_statements = []
        if not isinstance(original_node.body, cst.EmptyLine):
            for stmt in original_node.body.body:
                if (
                    hasattr(stmt, "body")
                    and not isinstance(stmt.body, cst.IndentedBlock)
                    and len(stmt.body) > 0
                    and isinstance(stmt.body[0], cst.Pass)
                ):
                    continue
                # Add all old class body except any `pass` statement
                class_body_statements.append(stmt)
        # Add new property to the class
        class_body_statements.append(
            cst.SimpleStatementLine(
                body=[
                    cst.AnnAssign(
                        target=self.property_name_node,
                        annotation=self.property_type_node,
                        value=self.property_value_node,
                    )
                ]
            )
        )
        new_body = cst.IndentedBlock(
            body=class_body_statements,
            indent=updated_node.body.indent,
        )
        return updated_node.with_changes(body=new_body)


@typechecked
def add_property_to_class(
    code: str,
    class_name: str,
    property_name: str,
    property_type: str,
    property_value: Optional[str] = None,
) -> str:
    module = cst.parse_module(code)
    property_name_node = cst.Name(value=property_name)
    property_type_node = _get_property_type_node(property_type)
    property_value_node = _get_property_value_node(property_value)
    transformed_module = module.visit(
        AddPropertyTransformer(
            class_name=class_name,
            property_name_node=property_name_node,
            property_type_node=property_type_node,
            property_value_node=property_value_node,
        )
    )
    return transformed_module.code


def _get_property_type_node(property_type: str) -> cst.Annotation:
    if property_type.startswith("Optional[") and property_type.endswith("]"):
        inner_type = property_type[len("Optional[") : -1]
        return cst.Annotation(
            annotation=cst.Subscript(
                value=cst.Name("Optional"),
                slice=[
                    cst.SubscriptElement(
                        slice=cst.Index(value=cst.Name(value=inner_type))
                    )
                ],
            )
        )
    else:
        return cst.Annotation(cst.Name(value=property_type))


def _get_property_value_node(
    property_value: Optional[str],
) -> Optional[cst.BaseExpression]:
    if property_value is None:
        return None
    return cst.parse_expression(property_value)



from config import APP_ENABLE_SNAKE_ZRB_MODULE_NAME_MODULE
from helper.migration import migrate
from integration.db_connection import engine
from integration.log import logger
from module.snake_zrb_module_name.integration import Base


async def migrate_snake_zrb_module_name():
    if not APP_ENABLE_SNAKE_ZRB_MODULE_NAME_MODULE:
        logger.info('🥪 Skip DB migration for "snake_zrb_module_name"')
        return
    logger.info('🥪 Perform DB migration for "snake_zrb_module_name"')
    await migrate(engine=engine, Base=Base)



import os

PROJECT_DIR = os.path.dirname(os.path.dirname(os.path.dirname(__file__)))
RESOURCE_DIR = os.path.join(PROJECT_DIR, "src", "kebab-zrb-app-name")
DEPLOYMENT_DIR = os.path.join(RESOURCE_DIR, "deployment")
DEPLOYMENT_TEMPLATE_ENV_FILE_NAME = os.path.join(DEPLOYMENT_DIR, "template.env")
APP_DIR = os.path.join(RESOURCE_DIR, "src")
APP_TEMPLATE_ENV_FILE_NAME = os.path.join(APP_DIR, "template.env")



from ._group import snake_zrb_app_name_microservices_group
from .deploy import deploy_snake_zrb_app_name_microservices
from .start import start_snake_zrb_app_name_microservices

assert snake_zrb_app_name_microservices_group
assert deploy_snake_zrb_app_name_microservices
assert start_snake_zrb_app_name_microservices



from zrb.builtin.docker._group import docker_group
from zrb.runner import runner
from zrb.task.cmd_task import CmdTask
from zrb.task_input.bool_input import BoolInput

prune_docker = CmdTask(
    name="prune",
    group=docker_group,
    description="Prune unused images and volumes",
    inputs=[
        BoolInput(
            name="all", shortcut="a", prompt="Remove all unused images", default=False
        ),
        BoolInput(
            name="volume", shortcut="v", prompt="Prune anonymous volume", default=False
        ),
    ],
    cmd=[
        "docker system prune -f {% if input.all %}-a{% endif %} {% if input.volume %}--volumes{% endif %}",  # noqa
        "docker image prune -f {% if input.all %}-a{% endif %}",
        "docker system df",
    ],
)
runner.register(prune_docker)



import os

import jsons

from zrb.helper.string.conversion import to_boolean

PREFER_MICROSERVICES = to_boolean(os.getenv("PROJECT_PREFER_MICROSERVICES", "0"))

AUTOMATE_DIR = os.path.dirname(__file__)
PROJECT_DIR = os.path.dirname(os.path.dirname(AUTOMATE_DIR))
RESOURCE_DIR = os.path.join(PROJECT_DIR, "src", "kebab-zrb-app-name")
DEPLOYMENT_DIR = os.path.join(RESOURCE_DIR, "deployment")
DEPLOYMENT_TEMPLATE_ENV_FILE_NAME = os.path.join(DEPLOYMENT_DIR, "template.env")
APP_DIR = os.path.join(RESOURCE_DIR, "src")
APP_FRONTEND_DIR = os.path.join(APP_DIR, "frontend")
APP_FRONTEND_BUILD_DIR = os.path.join(APP_FRONTEND_DIR, "build")
APP_TEMPLATE_ENV_FILE_NAME = os.path.join(APP_DIR, "template.env")
LOAD_TEST_DIR = os.path.join(RESOURCE_DIR, "loadtest")
LOAD_TEST_TEMPLATE_ENV_FILE_NAME = os.path.join(RESOURCE_DIR, "template.env")

MODULE_CONFIG_PATH = os.path.join(AUTOMATE_DIR, "config", "modules.json")

with open(MODULE_CONFIG_PATH) as file:
    MODULE_JSON_STR = file.read()
MODULES = jsons.loads(MODULE_JSON_STR)



from collections.abc import Mapping
from logging import Logger
from typing import Any, Union

from component.messagebus import Publisher
from component.repo import SearchFilter
from component.rpc import Caller, Server
from module.auth.integration.model.user_model import user_model
from module.auth.schema.token import AccessTokenData
from module.auth.schema.user import UserData, UserLogin


def register_rpc(
    logger: Logger, rpc_server: Server, rpc_caller: Caller, publisher: Publisher
):
    logger.info('🥪 Register RPC handlers for "auth.user"')

    @rpc_server.register("auth_user_is_admin")
    async def user_is_admin(id: str) -> bool:
        """
        Used by RPC Authenticator
        """
        return await user_model.is_admin(id)

    @rpc_server.register("auth_user_is_guest")
    async def user_is_guest(id: str) -> bool:
        """
        Used by RPC Authenticator
        """
        return await user_model.is_guest(id)

    @rpc_server.register("auth_is_user_authorized")
    async def is_user_having_permission(
        id: str, permission_name: Union[str, list[str]]
    ) -> Mapping[str, bool]:
        """
        Used by RPC Authenticator
        """
        if isinstance(permission_name, str):
            return await user_model.is_authorized(id, permission_name)
        return await user_model.is_authorized(id, *permission_name)

    @rpc_server.register("auth_create_token")
    async def create_token(login_data: Mapping[str, str]) -> Mapping[str, str]:
        result = await user_model.create_auth_token(UserLogin(**login_data))
        return result.model_dump()

    @rpc_server.register("auth_refresh_token")
    async def refresh_token(refresh_token: str, access_token: str) -> Mapping[str, str]:
        result = await user_model.refresh_auth_token(refresh_token, access_token)
        return result.model_dump()

    @rpc_server.register("auth_get_user")
    async def get(
        keyword: str,
        criterion: Mapping[str, Any],
        limit: int,
        offset: int,
        user_token_data: Mapping[str, Any],
    ) -> Mapping[str, Any]:
        result = await user_model.get(
            search_filter=SearchFilter(keyword=keyword, criterion=criterion),
            limit=limit,
            offset=offset,
        )
        return result.model_dump()

    @rpc_server.register("auth_get_user_by_id")
    async def get_by_id(
        id: str, user_token_data: Mapping[str, Any] = {}
    ) -> Mapping[str, Any]:
        row = await user_model.get_by_id(id)
        return row.model_dump()

    @rpc_server.register("auth_insert_user")
    async def insert(
        data: Mapping[str, Any], user_token_data: Mapping[str, Any]
    ) -> Mapping[str, Any]:
        user_token_data = AccessTokenData(**user_token_data)
        data["created_by"] = user_token_data.user_id
        data["updated_by"] = user_token_data.user_id
        row = await user_model.insert(data=UserData(**data))
        return row.model_dump()

    @rpc_server.register("auth_update_user")
    async def update(
        id: str, data: Mapping[str, Any], user_token_data: Mapping[str, Any]
    ) -> Mapping[str, Any]:
        user_token_data = AccessTokenData(**user_token_data)
        data["updated_by"] = user_token_data.user_id
        row = await user_model.update(id=id, data=UserData(**data))
        return row.model_dump()

    @rpc_server.register("auth_delete_user")
    async def delete(id: str, user_token_data: Mapping[str, Any]) -> Mapping[str, Any]:
        user_token_data = AccessTokenData(**user_token_data)
        row = await user_model.delete(id=id)
        return row.model_dump()



import os

from zrb.helper.codemod.append_code_to_function import append_code_to_function
from zrb.helper.file.text import read_text_file_async, write_text_file_async
from zrb.helper.typecheck import typechecked
from zrb.helper.util import to_snake_case
from zrb.task.task import Task

from ._common import get_app_module_dir


@typechecked
async def register_permission(
    task: Task, project_dir: str, app_name: str, module_name: str, entity_name: str
):
    snake_module_name = to_snake_case(module_name)
    snake_entity_name = to_snake_case(entity_name)
    module_register_permission_file_path = os.path.join(
        get_app_module_dir(project_dir, app_name),
        "auth",
        "register_permission.py",
    )
    task.print_out(f"Read code from: {module_register_permission_file_path}")
    code = await read_text_file_async(module_register_permission_file_path)
    code = append_code_to_function(
        code=code,
        function_name="register_permission",
        new_code="\n".join(
            [
                "await ensure_entity_permission(",
                f"    module_name='{snake_module_name}', entity_name='{snake_entity_name}'",  # noqa
                ")",
            ]
        ),
    )
    task.print_out(
        f'Add "ensure_entity_permission" call for {snake_entity_name} ' + "to the code"
    )
    task.print_out(f"Write modified code to: {module_register_permission_file_path}")
    await write_text_file_async(module_register_permission_file_path, code)



from typing import Any

from zrb import Task, python_task, runner
from zrb.builtin import project_group
from zrb.helper.accessories.color import colored


@python_task(
    name="deploy", group=project_group, description="Deploy project", runner=runner
)
def deploy_project(*args: Any, **kwargs: Any):
    task: Task = kwargs.get("_task")
    task.print_out(colored("Project deployed", color="yellow"))



import asyncio
import inspect
from collections.abc import Callable
from typing import Any

from zrb.helper.accessories.color import colored
from zrb.helper.log import logger
from zrb.helper.typecheck import typechecked

logger.debug(colored("Loading zrb.helper.callable", attrs=["dark"]))


@typechecked
async def run_async(fn: Callable, *args: Any, **kwargs: Any) -> Any:
    if inspect.iscoroutinefunction(fn):
        return await fn(*args, **kwargs)
    coro = asyncio.to_thread(fn, *args, **kwargs)
    task = asyncio.create_task(coro)
    return await task



import hashlib
from typing import Any

from zrb.builtin.md5._group import md5_group
from zrb.runner import runner
from zrb.task.decorator import python_task
from zrb.task_input.str_input import StrInput


@python_task(
    name="hash",
    group=md5_group,
    inputs=[StrInput(name="text", shortcut="t", description="Text", default="")],
    description="Hash md5",
    retry=0,
    runner=runner,
)
async def hash_text_md5(*args: str, **kwargs: Any):
    text: str = kwargs.get("text", "")
    hashed_text = hashlib.md5(text.encode()).hexdigest()
    return hashed_text



import os

from zrb.builtin.explain._group import explain_group
from zrb.runner import runner
from zrb.task.wiki_task import create_wiki_tasks

CURRENT_DIRECTORY = os.path.dirname(__file__)

explain_tasks = create_wiki_tasks(
    directory=os.path.join(CURRENT_DIRECTORY, "explanation"),
    group=explain_group,
    runner=runner,
)



from typing import Any

from zrb import Task, python_task, runner
from zrb.builtin import project_group


@python_task(
    name="kebab-zrb-task-name",
    description="human readable zrb task name",
    group=project_group,
    runner=runner,
)
async def snake_zrb_task_name(*args: Any, **kwargs: Any) -> Any:
    task: Task = kwargs.get("_task")
    env_map: Mapping[str, str] = task.get_env_map()
    input_map: Mapping[str, str] = task.get_input_map()
    task.print_out(f"Env map: {env_map}")
    task.print_out(f"Input map: {input_map}")
    return "ok"



import os

from zrb.helper.string.parse_replacement import parse_replacement
from zrb.helper.util import to_kebab_case, to_snake_case


def replace_marker(text: str, marker: str, code: str) -> str:
    return parse_replacement(text, {marker: "\n".join([code, marker])})


def get_app_frontend_routes_dir(project_dir: str, app_name: str) -> str:
    return os.path.join(
        get_app_dir(project_dir, app_name), "src", "frontend", "src", "routes"
    )


def get_app_dir(project_dir: str, app_name: str) -> str:
    kebab_app_name = to_kebab_case(app_name)
    return os.path.join(project_dir, "src", kebab_app_name)


def get_sqlalchemy_column_type(column_type: str) -> str:
    default_value = "String"
    type_map = {
        "string": default_value,
        "text": "Text",
        "boolean": "Boolean",
        "integer": "Integer",
        "float": "Float",
        "double": "Double",
        "date": "Date",
        "datetime": "DateTime",
        "time": "Time",
    }
    column_type = column_type.lower()
    return type_map.get(column_type.lower(), default_value)


def get_python_column_type(column_type: str) -> str:
    default_value = "str"
    type_map = {
        "string": default_value,
        "text": "str",
        "boolean": "bool",
        "integer": "int",
        "float": "float",
        "double": "double",
        "date": "date",
        "datetime": "datetime",
        "time": "time",
    }
    return type_map.get(column_type.lower(), default_value)


def get_python_value_for_testing(column_type: str) -> str:
    default_value = '"A string"'
    type_map = {
        "string": default_value,
        "text": '"A text"',
        "boolean": "True",
        "integer": "42",
        "float": "3.14",
        "double": "3.14",
        "date": "date(2024, 8, 10)",
        "datetime": "datetime(2024, 8, 10, 14, 30)",
        "time": "time(14, 30)",
    }
    return type_map.get(column_type.lower(), default_value)


def get_default_js_value(column_type: str) -> str:
    default_value = '""'
    type_map = {
        "string": default_value,
        "text": '""',
        "boolean": "true",
        "integer": "0",
        "float": "0.0",
        "double": "0.0",
        "date": "new Date()",
        "datetime": "new Date()",
        "time": "new Date().toLocaleTimeString('en-US', { hour12: false })",
    }
    return type_map.get(column_type.lower(), default_value)


def get_html_input(column_type: str, column_name: str, column_caption: str) -> str:
    snake_column_name = to_snake_case(column_name)
    kebab_column_name = to_kebab_case(column_name)
    default_value = f'<input type="text" class="input w-full" id="{kebab_column_name}" placeholder="{column_caption}" bind:value="{{row.{snake_column_name}}}" />'  # noqa
    type_map = {
        "string": default_value,
        "text": f'<textarea class="textarea w-full" id="{kebab_column_name}" placeholder="{column_caption}" bind:value="{{row.{snake_column_name}}}"></textarea>',  # noqa
        "boolean": f'<input type="checkbox" class="checkbox" id="{kebab_column_name}" bind:checked="{{row.{snake_column_name}}}" />',  # noqa
        "integer": f'<input type="number" class="input w-full" id="{kebab_column_name}" placeholder="{column_caption}" bind:value="{{row.{snake_column_name}}}" />',  # noqa
        "float": f'<input type="number" step="0.01" class="input w-full" id="{kebab_column_name}" placeholder="{column_caption}" bind:value="{{row.{snake_column_name}}}" />',  # noqa
        "double": f'<input type="number" step="0.01" class="input w-full" id="{kebab_column_name}" placeholder="{column_caption}" bind:value="{{row.{snake_column_name}}}" />',  # noqa
        "date": f'<input type="date" class="input w-full" id="{kebab_column_name}" placeholder="{column_caption}" bind:value="{{row.{snake_column_name}}}" />',  # noqa
        "datetime": f'<input type="datetime-local" class="input w-full" id="{kebab_column_name}" placeholder="{column_caption}" bind:value="{{row.{snake_column_name}}}" />',  # noqa
        "time": f'<input type="time" class="input w-full" id="{kebab_column_name}" placeholder="{column_caption}" bind:value="{{row.{snake_column_name}}}" />',  # noqa
    }
    return type_map.get(column_type.lower(), default_value)



from zrb import DockerComposeStartTask, runner

from ..._checker import (
    kafka_outside_checker,
    kafka_plaintext_checker,
    pandaproxy_outside_checker,
    pandaproxy_plaintext_checker,
    rabbitmq_checker,
    rabbitmq_management_checker,
    redpanda_console_checker,
)
from ..._constant import RESOURCE_DIR
from ..._input import host_input, https_input, local_input
from ...image._input import image_input
from .._env import compose_env_file
from .._input import enable_monitoring_input
from .._service_config import snake_zrb_app_name_service_configs
from ..remove import remove_snake_zrb_app_name_container
from ._group import snake_zrb_app_name_support_container_group
from ._helper import activate_support_compose_profile, should_start_support_container

start_snake_zrb_app_name_support_container = DockerComposeStartTask(
    icon="🐳",
    name="start",
    description="Start human readable zrb app name container",
    group=snake_zrb_app_name_support_container_group,
    inputs=[
        local_input,
        enable_monitoring_input,
        host_input,
        https_input,
        image_input,
    ],
    should_execute=should_start_support_container,
    upstreams=[remove_snake_zrb_app_name_container],
    cwd=RESOURCE_DIR,
    setup_cmd=activate_support_compose_profile,
    compose_env_prefix="CONTAINER_ZRB_ENV_PREFIX",
    compose_service_configs=snake_zrb_app_name_service_configs,
    env_files=[compose_env_file],
    checkers=[
        rabbitmq_checker,
        rabbitmq_management_checker,
        kafka_outside_checker,
        kafka_plaintext_checker,
        redpanda_console_checker,
        pandaproxy_outside_checker,
        pandaproxy_plaintext_checker,
    ],
)

runner.register(start_snake_zrb_app_name_support_container)



import os
from collections.abc import Mapping

from zrb import Env, EnvFile, ServiceConfig
from zrb.helper.util import to_kebab_case

from .._constant import APP_TEMPLATE_ENV_FILE_NAME, MODULES, RESOURCE_DIR

_otel_exporter_endpoint_env = Env(
    name="APP_OTEL_EXPORTER_OTLP_ENDPOINT",
    os_name="",
    default="http://otel-collector:4317",
)

_container_env_prefix = "CONTAINER_ZRB_ENV_PREFIX"
_app_template_env_file = EnvFile(
    path=APP_TEMPLATE_ENV_FILE_NAME, prefix=_container_env_prefix
)
_docker_compose_app_env_file_name = os.path.join(RESOURCE_DIR, "docker-compose-app.env")
_docker_compose_app_env_file = EnvFile(
    path=_docker_compose_app_env_file_name, prefix=_container_env_prefix
)
_disable_all_module_env_file_name = os.path.join(
    RESOURCE_DIR, "all-module-disabled.env"
)
_disable_all_module_env_file = EnvFile(
    path=_disable_all_module_env_file_name, prefix=_container_env_prefix
)
_enable_all_module_env_file_name = os.path.join(RESOURCE_DIR, "all-module-enabled.env")
_enable_all_module_env_file = EnvFile(
    path=_enable_all_module_env_file_name, prefix=_container_env_prefix
)

snake_zrb_app_name_service_configs: Mapping[str, ServiceConfig] = {
    "kebab-zrb-app-name": ServiceConfig(
        env_files=[
            _app_template_env_file,
            _docker_compose_app_env_file,
            _enable_all_module_env_file,
        ],
        envs=[_otel_exporter_endpoint_env],
    ),
    "kebab-zrb-app-name-gateway": ServiceConfig(
        env_files=[
            _app_template_env_file,
            _docker_compose_app_env_file,
            _enable_all_module_env_file,
        ],
        envs=[_otel_exporter_endpoint_env],
    ),
}
for module in MODULES:
    service_name = f"kebab-zrb-app-name-{to_kebab_case(module)}-service"
    snake_zrb_app_name_service_configs[service_name] = ServiceConfig(
        env_files=[
            _app_template_env_file,
            _docker_compose_app_env_file,
            _disable_all_module_env_file,
        ],
        envs=[_otel_exporter_endpoint_env],
    )



import asyncio
import os
from typing import Any

from zrb.builtin.project._helper import validate_existing_project_dir
from zrb.builtin.project._input import project_dir_input
from zrb.builtin.project.add.fastapp._group import project_add_fastapp_group
from zrb.builtin.project.add.fastapp.app._input import app_name_input
from zrb.builtin.project.add.fastapp.crud._input import entity_name_input
from zrb.builtin.project.add.fastapp.field._helper import (
    inject_delete_page,
    inject_detail_page,
    inject_insert_page,
    inject_list_page,
    inject_repo,
    inject_schema,
    inject_test,
    inject_update_page,
)
from zrb.builtin.project.add.fastapp.field._input import (
    column_name_input,
    column_type_input,
)
from zrb.builtin.project.add.fastapp.module._input import module_name_input
from zrb.helper import util
from zrb.helper.accessories.color import colored
from zrb.runner import runner
from zrb.task.decorator import python_task
from zrb.task.task import Task


@python_task(
    name="validate",
    inputs=[project_dir_input, app_name_input, module_name_input, entity_name_input],
    retry=0,
)
async def validate(*args: Any, **kwargs: Any):
    project_dir = kwargs.get("project_dir")
    validate_existing_project_dir(project_dir)
    app_name = kwargs.get("app_name")
    module_name = kwargs.get("module_name")
    entity_name = kwargs.get("entity_name")
    snake_module_name = util.to_snake_case(module_name)
    snake_entity_name = util.to_snake_case(entity_name)
    automation_dir = os.path.join(
        project_dir, "_automate", util.to_snake_case(app_name)
    )
    if not os.path.exists(automation_dir):
        raise Exception(f"Automation directory does not exist: {automation_dir}")
    app_dir = os.path.join(project_dir, "src", f"{util.to_kebab_case(app_name)}")
    if not os.path.exists(app_dir):
        raise Exception(f"App directory does not exist: {app_dir}")
    module_path = os.path.join(app_dir, "src", "module", snake_module_name)
    if not os.path.exists(module_path):
        raise Exception(f"Module directory does not exist: {module_path}")
    entity_path = os.path.join(module_path, "entity", snake_entity_name)
    if not os.path.exists(entity_path):
        raise Exception(f"Entity directory does not exist: {entity_path}")


@python_task(
    name="field",
    group=project_add_fastapp_group,
    description="Add Fastapp Field",
    inputs=[
        project_dir_input,
        app_name_input,
        module_name_input,
        entity_name_input,
        column_name_input,
        column_type_input,
    ],
    upstreams=[validate],
    runner=runner,
)
async def add_fastapp_field(*args: Any, **kwargs: Any):
    task: Task = kwargs.get("_task")
    project_dir = kwargs.get("project_dir", ".")
    app_name = kwargs.get("app_name")
    module_name = kwargs.get("module_name")
    entity_name = kwargs.get("entity_name")
    column_name = kwargs.get("column_name")
    column_type = kwargs.get("column_type")
    await asyncio.gather(
        asyncio.create_task(
            inject_test(
                task=task,
                project_dir=project_dir,
                app_name=app_name,
                module_name=module_name,
                entity_name=entity_name,
                column_name=column_name,
                column_type=column_type,
            )
        ),
        asyncio.create_task(
            inject_schema(
                task=task,
                project_dir=project_dir,
                app_name=app_name,
                module_name=module_name,
                entity_name=entity_name,
                column_name=column_name,
                column_type=column_type,
            )
        ),
        asyncio.create_task(
            inject_repo(
                task=task,
                project_dir=project_dir,
                app_name=app_name,
                module_name=module_name,
                entity_name=entity_name,
                column_name=column_name,
                column_type=column_type,
            )
        ),
        asyncio.create_task(
            inject_list_page(
                task=task,
                project_dir=project_dir,
                app_name=app_name,
                module_name=module_name,
                entity_name=entity_name,
                column_name=column_name,
            )
        ),
        asyncio.create_task(
            inject_detail_page(
                task=task,
                project_dir=project_dir,
                app_name=app_name,
                module_name=module_name,
                entity_name=entity_name,
                column_name=column_name,
            )
        ),
        asyncio.create_task(
            inject_delete_page(
                task=task,
                project_dir=project_dir,
                app_name=app_name,
                module_name=module_name,
                entity_name=entity_name,
                column_name=column_name,
            )
        ),
        asyncio.create_task(
            inject_update_page(
                task=task,
                project_dir=project_dir,
                app_name=app_name,
                module_name=module_name,
                entity_name=entity_name,
                column_name=column_name,
                column_type=column_type,
            )
        ),
        asyncio.create_task(
            inject_insert_page(
                task=task,
                project_dir=project_dir,
                app_name=app_name,
                module_name=module_name,
                entity_name=entity_name,
                column_name=column_name,
                column_type=column_type,
            )
        ),
    )
    task.print_out(colored("Fastapp crud added", color="yellow"))



from ._group import snake_zrb_app_name_image_group
from .build import build_snake_zrb_app_name_image
from .push import push_snake_zrb_app_name_image

assert snake_zrb_app_name_image_group
assert build_snake_zrb_app_name_image
assert push_snake_zrb_app_name_image



from zrb.builtin.devtool._group import devtool_group
from zrb.task_group.group import Group

devtool_install_group = Group(
    name="install", description="Install developer tools", parent=devtool_group
)



from _common import NAMESPACE, RABBITMQ_AUTH_PASSWORD, RABBITMQ_AUTH_USERNAME
from pulumi_kubernetes.helm.v3 import Chart, LocalChartOpts


def create_rabbitmq() -> Chart:
    rabbitmq = Chart(
        "rabbitmq",
        LocalChartOpts(
            path="./helm-charts/rabbitmq",
            namespace=NAMESPACE,
            values={
                "auth": {
                    "username": RABBITMQ_AUTH_USERNAME,
                    "password": RABBITMQ_AUTH_PASSWORD,
                },
                "resources": {
                    "limits": {"cpu": "1000m", "memory": "2Gi"},
                    "requests": {"cpu": "1000m", "memory": "1Gi"},
                },
            },
            skip_await=True,
        ),
    )
    return rabbitmq



from abc import ABC
from collections.abc import Mapping
from typing import Any

from component.repo import DBEntityMixin, DBRepo, Repo
from module.auth.entity.permission.repo import DBEntityPermission
from module.auth.entity.table import group_permission
from module.auth.integration import Base
from module.auth.schema.group import Group, GroupData
from sqlalchemy import Column, String
from sqlalchemy.orm import Session, relationship


class DBEntityGroup(Base, DBEntityMixin):
    class Config:
        from_attributes = True

    __tablename__ = "groups"
    name = Column(String)
    description = Column(String)
    permissions = relationship("DBEntityPermission", secondary=group_permission)


class GroupRepo(Repo[Group, GroupData], ABC):
    pass


class GroupDBRepo(DBRepo[Group, GroupData], GroupRepo):
    schema_cls = Group
    db_entity_cls = DBEntityGroup

    def _schema_data_to_db_entity_map(
        self, db: Session, group_data: GroupData
    ) -> Mapping[str, Any]:
        db_entity_map = super()._schema_data_to_db_entity_map(db, group_data)
        # Transform permissions
        db_entity_map["permissions"] = (
            db.query(DBEntityPermission)
            .filter(DBEntityPermission.id.in_(group_data.permissions))
            .all()
        )
        return db_entity_map



from component.repo.db_entity_mixin import DBEntityMixin
from component.repo.db_repo import DBRepo
from component.repo.repo import Repo
from component.repo.search_filter import SearchFilter

assert Repo
assert DBRepo
assert DBEntityMixin
assert SearchFilter



from zrb.builtin.project.add.task._group import project_add_task_group
from zrb.builtin.project.add.task.cmd import add_cmd_task
from zrb.builtin.project.add.task.docker_compose import add_docker_compose_task
from zrb.builtin.project.add.task.python import add_python_task

assert project_add_task_group
assert add_cmd_task
assert add_docker_compose_task
assert add_python_task



from zrb.task_input.choice_input import ChoiceInput
from zrb.task_input.str_input import StrInput

column_name_input = StrInput(
    name="column-name",
    shortcut="c",
    description="Column name",
    prompt="Column name",
    default="title",
)

column_type_input = ChoiceInput(
    name="column-type",
    shortcut="t",
    description="Column type",
    prompt="Column type",
    choices=[
        "string",
        "text",
        "boolean",
        "integer",
        "float",
        "double",
        "date",
        "datetime",
        "time",
    ],
    default="string",
)



import hashlib
from typing import Any

import aiofiles

from zrb.builtin.md5._group import md5_group
from zrb.runner import runner
from zrb.task.decorator import python_task
from zrb.task_input.str_input import StrInput


@python_task(
    name="sum",
    group=md5_group,
    inputs=[StrInput(name="file", shortcut="f", description="File", default="")],
    description="Sum md5 file",
    retry=0,
    runner=runner,
)
async def sum_file_md5(*args: str, **kwargs: Any):
    file_path: str = kwargs.get("file", "")
    async with aiofiles.open(file_path, mode="rb") as file:
        contents = await file.read()
    hashed_text = hashlib.md5(contents).hexdigest()
    return hashed_text



from typing import TypeVar

from zrb.config.config import ENABLE_TYPE_CHECKING
from zrb.helper.accessories.untyped_color import untyped_colored as colored
from zrb.helper.log import logger

logger.debug(colored("Loading zrb.helper.typecheck", attrs=["dark"]))
T = TypeVar("T")


def typechecked(anything: T) -> T:
    if ENABLE_TYPE_CHECKING:
        from beartype import beartype

        return beartype(anything)
    return anything



import os
from typing import Any

from zrb.builtin.project._helper import (
    create_register_module,
    validate_existing_project_dir,
    validate_inexisting_automation,
)
from zrb.builtin.project._input import project_dir_input
from zrb.builtin.project.add.task._group import project_add_task_group
from zrb.builtin.project.add.task._input import task_name_input
from zrb.builtin.project.add.task.docker_compose._input import (
    compose_command_input,
    http_port_input,
)
from zrb.runner import runner
from zrb.task.decorator import python_task
from zrb.task.resource_maker import ResourceMaker
from zrb.task.task import Task

_CURRENT_DIR = os.path.dirname(__file__)


@python_task(
    name="validate",
    inputs=[
        project_dir_input,
        task_name_input,
    ],
    retry=0,
)
async def validate(*args: Any, **kwargs: Any):
    project_dir = kwargs.get("project_dir")
    validate_existing_project_dir(project_dir)
    task_name = kwargs.get("task_name")
    validate_inexisting_automation(project_dir, task_name)


copy_resource = ResourceMaker(
    name="copy-resource",
    inputs=[
        project_dir_input,
        task_name_input,
        compose_command_input,
        http_port_input,
    ],
    upstreams=[validate],
    replacements={
        "zrbTaskName": "{{input.task_name}}",
        "zrbHttpPort": "{{input.http_port}}",
        "zrbComposeCommand": "{{input.compose_command}}",
    },
    template_path=os.path.join(_CURRENT_DIR, "template"),
    destination_path="{{ input.project_dir }}",
    excludes=[
        "*/__pycache__",
    ],
)

register_module = create_register_module(
    module_path="_automate.{{util.to_snake_case(input.task_name)}}",
    alias="_automate_{{util.to_snake_case(input.task_name)}}",
    inputs=[task_name_input],
    upstreams=[copy_resource],
)


@python_task(
    name="docker-compose",
    group=project_add_task_group,
    inputs=[
        project_dir_input,
        task_name_input,
        compose_command_input,
        http_port_input,
    ],
    upstreams=[register_module],
    runner=runner,
)
async def add_docker_compose_task(*args: Any, **kwargs: Any):
    task: Task = kwargs.get("_task")
    task.print_out("Success")



from zrb.builtin.ubuntu.install._group import ubuntu_install_group
from zrb.builtin.ubuntu.update import update_ubuntu
from zrb.runner import runner
from zrb.task.cmd_task import CmdTask

install_ubuntu_essentials = CmdTask(
    name="essentials",
    group=ubuntu_install_group,
    description="Install ubuntu essential packages",
    cmd=[
        "sudo apt install -y \\",
        "build-essential python3-distutils libssl-dev zlib1g-dev \\"
        "libbz2-dev libreadline-dev libsqlite3-dev libpq-dev python3-dev \\",
        "llvm libncurses5-dev libncursesw5-dev xz-utils tk-dev libffi-dev \\",
        "liblzma-dev python3-openssl libblas-dev liblapack-dev rustc \\",
        "golang gfortran fd-find ripgrep wget curl git ncat zip unzip \\",
        "cmake make tree tmux zsh neovim xdotool xsel",
    ],
    retry_interval=3,
    preexec_fn=None,
)
update_and_install_essentials: CmdTask = install_ubuntu_essentials.copy()
update_and_install_essentials.add_upstream(update_ubuntu)
runner.register(update_and_install_essentials)



import os
from typing import Any

from zrb.builtin.project._helper import (
    create_register_module,
    validate_existing_project_dir,
    validate_inexisting_automation,
)
from zrb.builtin.project._input import project_dir_input
from zrb.builtin.project.add._group import project_add_group
from zrb.builtin.project.add.plugin._input import (
    package_author_email_input,
    package_author_name_input,
    package_description_input,
    package_documentation_input,
    package_homepage_input,
    package_name_input,
    package_repository_input,
)
from zrb.builtin.project.add.project_task import add_project_tasks
from zrb.config.config import VERSION
from zrb.helper.accessories.color import colored
from zrb.runner import runner
from zrb.task.decorator import python_task
from zrb.task.resource_maker import ResourceMaker
from zrb.task.task import Task

_CURRENT_DIR = os.path.dirname(__file__)


@python_task(
    name="validate",
    inputs=[project_dir_input, package_name_input],
    retry=0,
)
async def validate(*args: Any, **kwargs: Any):
    project_dir = kwargs.get("project_dir")
    validate_existing_project_dir(project_dir)
    package_name = kwargs.get("package_name")
    validate_inexisting_automation(project_dir, package_name)


copy_resource = ResourceMaker(
    name="copy-resource",
    inputs=[
        project_dir_input,
        package_name_input,
        package_description_input,
        package_homepage_input,
        package_repository_input,
        package_documentation_input,
        package_author_name_input,
        package_author_email_input,
    ],
    upstreams=[validate],
    replacements={
        "zrbPackageName": "{{input.package_name}}",
        "zrbPackageDescription": "{{input.package_description}}",
        "zrbPackageHomepage": "{{input.package_homepage}}",
        "zrbPackageRepository": "{{input.package_repository}}",
        "zrbPackageDocumentation": "{{input.package_documentation}}",
        "zrbPackageAuthorName": "{{input.package_author_name}}",
        "zrbPackageAuthorEmail": "{{input.package_author_email}}",
        "zrbVersion": VERSION,
    },
    template_path=os.path.join(_CURRENT_DIR, "template"),
    destination_path="{{ input.project_dir }}",
    excludes=[
        "*/__pycache__",
    ],
)

register_plugin = create_register_module(
    module_path="_automate.{{util.to_snake_case(input.package_name)}}",
    alias="_automate_{{util.to_snake_case(input.package_name)}}",
    inputs=[package_name_input],
    upstreams=[copy_resource, add_project_tasks],
    task_name="register-plugin",
)


@python_task(
    name="plugin",
    group=project_add_group,
    description="Add plugin to project",
    upstreams=[register_plugin],
    inputs=[
        project_dir_input,
        package_name_input,
        package_description_input,
        package_homepage_input,
        package_repository_input,
        package_documentation_input,
        package_author_name_input,
        package_author_email_input,
    ],
    runner=runner,
)
async def add_plugin(*args: Any, **kwargs: Any):
    task: Task = kwargs.get("_task")
    task.print_out(colored("Plugin added", color="yellow"))



import sys
from collections.abc import Callable, Mapping
from typing import Any, Union

import click

from zrb.helper.accessories.color import colored
from zrb.helper.asyncio_task import stop_asyncio_sync
from zrb.helper.log import logger
from zrb.helper.typecheck import typechecked
from zrb.task.any_task import AnyTask
from zrb.task_group.group import Group as TaskGroup

logger.debug(colored("Loading zrb.action.runner", attrs=["dark"]))

CliSubcommand = Union[click.Group, click.Command]


@typechecked
class Runner:
    """
    Runner class.
    Any tasks registered to the runner will be accessible from the terminal
    """

    def __init__(self, env_prefix: str = ""):
        logger.info(colored("Create runner", attrs=["dark"]))
        self.__env_prefix = env_prefix
        self.__tasks: list[AnyTask] = []
        self.__registered_groups: Mapping[str, click.Group] = {}
        self.__top_levels: list[CliSubcommand] = []
        self.__subcommands: Mapping[str, list[click.Group]] = {}
        self.__registered_task_cli_name: list[str] = []
        logger.info(colored("Runner created", attrs=["dark"]))

    def register(self, *tasks: AnyTask):
        for task in tasks:
            task._set_has_cli_interface()
            cli_name = task._get_full_cli_name()
            if cli_name in self.__registered_task_cli_name:
                raise RuntimeError(f'Task "{cli_name}" has already been registered')
            logger.debug(colored(f'Register task: "{cli_name}"', attrs=["dark"]))
            self.__tasks.append(task)
            self.__registered_task_cli_name.append(cli_name)
            logger.debug(colored(f'Task registered: "{cli_name}"', attrs=["dark"]))

    def serve(self, cli: click.Group) -> click.Group:
        for task in self.__tasks:
            subcommand = self.__create_cli_subcommand(task)
            if subcommand not in self.__top_levels:
                self.__top_levels.append(subcommand)
                cli.add_command(subcommand)
        return cli

    def __create_cli_subcommand(
        self, task: AnyTask
    ) -> Union[click.Group, click.Command]:
        subcommand: CliSubcommand = self.__create_cli_command(task)
        task_group = task._group
        while task_group is not None:
            group = self.__register_sub_command(task_group, subcommand)
            parent_group = task_group.get_parent()
            if parent_group is None:
                return group
            subcommand = group
            task_group = parent_group
        return subcommand

    def __register_sub_command(
        self, task_group: TaskGroup, subcommand: CliSubcommand
    ) -> click.Group:
        task_group_id = task_group._get_full_cli_name()
        group = self.__get_cli_group(task_group)
        if task_group_id not in self.__subcommands:
            self.__subcommands[task_group_id] = []
        if subcommand not in self.__subcommands[task_group_id]:
            group.add_command(subcommand)
            self.__subcommands[task_group_id].append(subcommand)
        return group

    def __get_cli_group(self, task_group: TaskGroup) -> click.Group:
        task_group_id = task_group._get_full_cli_name()
        if task_group_id in self.__registered_groups:
            return self.__registered_groups[task_group_id]
        group_cli_name = task_group.get_cli_name()
        group_description = task_group.get_description()
        group = click.Group(name=group_cli_name, help=group_description)
        self.__registered_groups[task_group_id] = group
        return group

    def __create_cli_command(self, task: AnyTask) -> click.Command:
        task_inputs = task._get_combined_inputs()
        task_cli_name = task.get_cli_name()
        task_description = task.get_description()
        callback = self.__get_wrapped_task_function(task)
        command = click.Command(
            callback=callback, name=task_cli_name, help=task_description
        )
        # by default, add an argument named _args
        command.params.append(click.Argument(["_args"], nargs=-1))
        # add task inputs,
        # if there are inputs with the same name, choose the first.
        registered_input: Mapping[str, bool] = {}
        for task_input in task_inputs:
            if task_input.get_name() in registered_input:
                continue
            registered_input[task_input.get_name()] = True
            param_decl = task_input.get_param_decl()
            options = task_input.get_options()
            command.params.append(click.Option(param_decl, **options))
        return command

    def __get_wrapped_task_function(self, task: AnyTask) -> Callable[..., Any]:
        def wrapped_function(*args: Any, **kwargs: Any) -> Any:
            function = task.to_function(
                env_prefix=self.__env_prefix,
                raise_error=True,
                should_clear_xcom=True,
                should_stop_looper=True,
            )
            try:
                function(*args, **kwargs)
            except Exception:
                stop_asyncio_sync()
                sys.exit(1)
            finally:
                task.clear_xcom()

        return wrapped_function



from ._group import project_container_group
from .remove import remove_project_containers
from .start import start_project_containers
from .stop import stop_project_containers

assert project_container_group
assert remove_project_containers
assert start_project_containers
assert stop_project_containers



import os
import pathlib
from collections.abc import Callable, Iterable, Mapping
from typing import Any, Optional, TypeVar, Union

from zrb.helper.accessories.color import colored
from zrb.helper.log import logger
from zrb.helper.typecheck import typechecked
from zrb.helper.typing import JinjaTemplate
from zrb.helper.util import to_snake_case
from zrb.task.any_task import AnyTask
from zrb.task.any_task_event_handler import (
    OnFailed,
    OnReady,
    OnRetry,
    OnSkipped,
    OnStarted,
    OnTriggered,
    OnWaiting,
)
from zrb.task.base_task.base_task import BaseTask
from zrb.task.cmd_task import CmdTask, CmdVal
from zrb.task_env.env import Env
from zrb.task_env.env_file import EnvFile
from zrb.task_group.group import Group
from zrb.task_input.any_input import AnyInput

logger.debug(colored("Loading zrb.task.base_remote_cmd_task", attrs=["dark"]))

TSingleBaseRemoteCmdTask = TypeVar(
    "TSingleBaseRemoteCmdTask", bound="SingleBaseRemoteCmdTask"
)
TBaseRemoteCmdTask = TypeVar("TBaseRemoteCmdTask", bound="BaseRemoteCmdTask")


@typechecked
class RemoteConfig:
    def __init__(
        self,
        host: JinjaTemplate,
        user: JinjaTemplate = "",
        password: JinjaTemplate = "",
        ssh_key: JinjaTemplate = "",
        port: Union[int, JinjaTemplate] = 22,
        name: Optional[str] = None,
        config_map: Optional[Mapping[str, JinjaTemplate]] = None,
    ):
        self.name = name if name is not None else host
        self.host = host
        self.user = user
        self.password = password
        self.ssh_key = ssh_key
        self.port = port
        self.config_map = {} if config_map is None else config_map


@typechecked
class SingleBaseRemoteCmdTask(CmdTask):
    def __init__(
        self,
        name: str,
        remote_config: RemoteConfig,
        group: Optional[Group] = None,
        inputs: Iterable[AnyInput] = [],
        envs: Iterable[Env] = [],
        env_files: Iterable[EnvFile] = [],
        icon: Optional[str] = None,
        color: Optional[str] = None,
        description: str = "",
        executable: Optional[str] = None,
        pre_cmd: CmdVal = "",
        pre_cmd_path: CmdVal = "",
        cmd: CmdVal = "",
        cmd_path: CmdVal = "",
        post_cmd: CmdVal = "",
        post_cmd_path: CmdVal = "",
        cwd: Optional[Union[str, pathlib.Path]] = None,
        should_render_cwd: bool = True,
        upstreams: Iterable[AnyTask] = [],
        fallbacks: Iterable[AnyTask] = [],
        on_triggered: Optional[OnTriggered] = None,
        on_waiting: Optional[OnWaiting] = None,
        on_skipped: Optional[OnSkipped] = None,
        on_started: Optional[OnStarted] = None,
        on_ready: Optional[OnReady] = None,
        on_retry: Optional[OnRetry] = None,
        on_failed: Optional[OnFailed] = None,
        checkers: Iterable[AnyTask] = [],
        checking_interval: Union[float, int] = 0.05,
        retry: int = 2,
        retry_interval: Union[float, int] = 1,
        max_output_line: int = 1000,
        max_error_line: int = 1000,
        preexec_fn: Optional[Callable[[], Any]] = os.setsid,
        should_execute: Union[bool, str, Callable[..., bool]] = True,
        return_upstream_result: bool = False,
        should_print_cmd_result: bool = True,
        should_show_cmd: bool = True,
        should_show_working_directory: bool = True,
    ):
        CmdTask.__init__(
            self,
            name=name,
            group=group,
            inputs=inputs,
            envs=envs,
            env_files=env_files,
            icon=icon,
            color=color,
            description=description,
            executable=executable,
            cmd=cmd,
            cmd_path=cmd_path,
            cwd=cwd,
            should_render_cwd=should_render_cwd,
            upstreams=upstreams,
            fallbacks=fallbacks,
            on_triggered=on_triggered,
            on_waiting=on_waiting,
            on_skipped=on_skipped,
            on_started=on_started,
            on_ready=on_ready,
            on_retry=on_retry,
            on_failed=on_failed,
            checkers=checkers,
            checking_interval=checking_interval,
            retry=retry,
            retry_interval=retry_interval,
            max_output_line=max_output_line,
            max_error_line=max_error_line,
            preexec_fn=preexec_fn,
            should_execute=should_execute,
            return_upstream_result=return_upstream_result,
            should_print_cmd_result=should_print_cmd_result,
            should_show_cmd=should_show_cmd,
            should_show_working_directory=should_show_working_directory,
        )
        self._pre_cmd = pre_cmd
        self._pre_cmd_path = pre_cmd_path
        self._post_cmd = post_cmd
        self._post_cmd_path = post_cmd_path
        self._remote_config = remote_config

    def copy(self) -> TSingleBaseRemoteCmdTask:
        return super().copy()

    def inject_envs(self):
        super().inject_envs()
        # add remote config properties as env
        self.add_env(
            Env(
                name="_CONFIG_HOST",
                os_name="",
                default=self.render_str(self._remote_config.host),
            ),
            Env(
                name="_CONFIG_PORT",
                os_name="",
                default=str(self.render_int(self._remote_config.port)),
            ),
            Env(
                name="_CONFIG_SSH_KEY",
                os_name="",
                default=self.render_str(self._remote_config.ssh_key),
            ),
            Env(
                name="_CONFIG_USER",
                os_name="",
                default=self.render_str(self._remote_config.user),
            ),
            Env(
                name="_CONFIG_PASSWORD",
                os_name="",
                default=self.render_str(self._remote_config.password),
            ),
        )
        for key, val in self._remote_config.config_map.items():
            upper_snake_key = to_snake_case(key).upper()
            rendered_val = self.render_str(val)
            # add remote config map as env
            self.add_env(
                Env(
                    name="_CONFIG_MAP_" + upper_snake_key,
                    os_name="",
                    default=rendered_val,
                )
            )

    def get_cmd_script(self, *args: Any, **kwargs: Any) -> str:
        cmd_str = "\n".join(
            [
                self._create_cmd_script(
                    self._pre_cmd_path, self._pre_cmd, *args, **kwargs
                ),
                super().get_cmd_script(*args, **kwargs),
                self._create_cmd_script(
                    self._post_cmd_path, self._post_cmd, *args, **kwargs
                ),
            ]
        )
        return cmd_str


@typechecked
class BaseRemoteCmdTask(BaseTask):
    def __init__(
        self,
        name: str,
        remote_configs: Iterable[RemoteConfig],
        group: Optional[Group] = None,
        inputs: Iterable[AnyInput] = [],
        envs: Iterable[Env] = [],
        env_files: Iterable[EnvFile] = [],
        icon: Optional[str] = None,
        color: Optional[str] = None,
        description: str = "",
        executable: Optional[str] = None,
        pre_cmd: CmdVal = "",
        pre_cmd_path: CmdVal = "",
        cmd: CmdVal = "",
        cmd_path: CmdVal = "",
        post_cmd: CmdVal = "",
        post_cmd_path: CmdVal = "",
        cwd: Optional[Union[JinjaTemplate, pathlib.Path]] = None,
        should_render_cwd: bool = True,
        upstreams: Iterable[AnyTask] = [],
        fallbacks: Iterable[AnyTask] = [],
        on_triggered: Optional[OnTriggered] = None,
        on_waiting: Optional[OnWaiting] = None,
        on_skipped: Optional[OnSkipped] = None,
        on_started: Optional[OnStarted] = None,
        on_ready: Optional[OnReady] = None,
        on_retry: Optional[OnRetry] = None,
        on_failed: Optional[OnFailed] = None,
        checkers: Iterable[AnyTask] = [],
        checking_interval: Union[float, int] = 0.05,
        retry: int = 2,
        retry_interval: Union[float, int] = 1,
        max_output_line: int = 1000,
        max_error_line: int = 1000,
        preexec_fn: Optional[Callable[[], Any]] = os.setsid,
        should_execute: Union[bool, str, Callable[..., bool]] = True,
        return_upstream_result: bool = False,
        should_print_cmd_result: bool = True,
        should_show_cmd: bool = True,
        should_show_working_directory: bool = True,
    ):
        self._remote_configs = list(remote_configs)
        self._sub_tasks = [
            SingleBaseRemoteCmdTask(
                name=f"{name}-{remote_config.name}",
                remote_config=remote_config,
                inputs=inputs,
                envs=envs,
                env_files=env_files,
                description=description,
                executable=executable,
                pre_cmd=pre_cmd,
                pre_cmd_path=pre_cmd_path,
                cmd=cmd,
                cmd_path=cmd_path,
                post_cmd=post_cmd,
                post_cmd_path=post_cmd_path,
                cwd=cwd,
                should_render_cwd=should_render_cwd,
                upstreams=upstreams,
                fallbacks=fallbacks,
                on_triggered=on_triggered,
                on_waiting=on_waiting,
                on_skipped=on_skipped,
                on_started=on_started,
                on_ready=on_ready,
                on_retry=on_retry,
                on_failed=on_failed,
                checkers=checkers,
                checking_interval=checking_interval,
                retry=retry,
                retry_interval=retry_interval,
                max_output_line=max_output_line,
                max_error_line=max_error_line,
                preexec_fn=preexec_fn,
                should_execute=should_execute,
                return_upstream_result=return_upstream_result,
                should_print_cmd_result=should_print_cmd_result,
                should_show_cmd=should_show_cmd,
                should_show_working_directory=should_show_working_directory,
            )
            for remote_config in self._remote_configs
        ]
        BaseTask.__init__(
            self,
            name=name,
            icon=icon,
            color=color,
            group=group,
            description=description,
            upstreams=self._sub_tasks,
            retry=0,
            return_upstream_result=True,
        )

    def insert_input(self, *inputs: AnyInput):
        super().insert_input(*inputs)
        for subtask in self._sub_tasks:
            subtask.insert_input(*inputs)

    def add_input(self, *inputs: AnyInput):
        super().add_input(*inputs)
        for subtask in self._sub_tasks:
            subtask.add_input(*inputs)

    def insert_env(self, *envs: Env):
        super().insert_env(*envs)
        for subtask in self._sub_tasks:
            subtask.insert_env(*envs)

    def add_env(self, *envs: Env):
        super().add_env(*envs)
        for subtask in self._sub_tasks:
            subtask.add_env(*envs)

    def insert_env_file(self, *env_files: EnvFile):
        super().insert_env_file(*env_files)
        for subtask in self._sub_tasks:
            subtask.insert_env_file(*env_files)

    def add_env_file(self, *env_files: Env):
        super().add_env_file(*env_files)
        for subtask in self._sub_tasks:
            subtask.add_env_file(*env_files)

    def insert_upstream(self, *upstreams: AnyTask):
        for subtask in self._sub_tasks:
            subtask.insert_upstream(*upstreams)

    def add_upstream(self, *upstreams: AnyTask):
        for subtask in self._sub_tasks:
            subtask.add_upstream(*upstreams)

    def insert_fallback(self, *fallbacks: AnyTask):
        for subtask in self._sub_tasks:
            subtask.insert_fallbacks(*fallbacks)

    def add_fallback(self, *fallbacks: AnyTask):
        for subtask in self._sub_tasks:
            subtask.add_fallback(*fallbacks)

    def insert_checker(self, *checkers: AnyTask):
        for subtask in self._sub_tasks:
            subtask.insert_checkers(*checkers)

    def add_checker(self, *checkers: AnyTask):
        for subtask in self._sub_tasks:
            subtask.add_checker(*checkers)



from typing import Optional

from dotenv import dotenv_values

from zrb.helper.accessories.color import colored
from zrb.helper.log import logger
from zrb.helper.string.modification import double_quote
from zrb.helper.typecheck import typechecked
from zrb.task_env.constant import RESERVED_ENV_NAMES
from zrb.task_env.env import Env

logger.debug(colored("Loading zrb.task_env.env_file", attrs=["dark"]))

# flake8: noqa E501


@typechecked
class EnvFile:
    """
    Env is an interface for a Task to read and use OS Environment File.

    Under the hood, EnvFile creates list of Env based on its `path` property.

    Attributes:
        path (str): The path to the environment file.
        prefix (Optional[str]): An optional prefix to be applied to environment variables.
        should_render (bool): Flag to determine if the environment values should be rendered.

    Examples:
        >>> from zrb import EnvFile, Task
        >>> import os
        >>> CURRENT_DIR = os.dirname(__file__)
        >>> task = Task(
        >>>     name='task',
        >>>     env_files=[
        >>>         EnvFile(path=os.path.join(CURRENT_DIR, '.env'), prefix='SYSTEM')
        >>>     ]
        >>> )
    """

    def __init__(
        self, path: str, prefix: Optional[str] = None, should_render: bool = False
    ):
        self.__path = path
        self.__prefix = prefix.upper() if prefix is not None else None
        self.__should_render = should_render
        self.__env_list: list[Env] = []
        self.__env_list_fetched: bool = False

    def get_envs(self) -> list[Env]:
        """
        Retrieves a list of Env objects based on the environment file. If a prefix is provided, it is
        applied to the environment variable names.

        Returns:
            list[Env]: A list of Env objects representing the environment variables defined in the file.

        Examples:
            >>> from zrb import Env, EnvFile
            >>> env_file = EnvFile(path='some_file.env')
            >>> envs: list[Env] = env_file.get_envs()
        """
        if self.__env_list_fetched:
            return self.__env_list
        env_list: list[Env] = []
        env_map = dotenv_values(self.__path)
        for key, value in env_map.items():
            if key in RESERVED_ENV_NAMES:
                continue
            os_name: Optional[str] = None
            if self.__prefix is not None and self.__prefix != "":
                os_name = f"{self.__prefix}_{key}"
            env_list.append(
                Env(
                    name=key,
                    os_name=os_name,
                    default=value,
                    should_render=self.__should_render,
                )
            )
        self.__env_list = env_list
        self.__env_list_fetched = True
        return env_list

    def __repr__(self) -> str:
        cls_name = self.__class__.__name__
        path = double_quote(self.__path)
        prefix = double_quote(self.__prefix)
        return f"<{cls_name} path={path} prefix={prefix}>"



from zrb.action.runner import Runner
from zrb.config.config import ENV_PREFIX
from zrb.helper.accessories.color import colored
from zrb.helper.log import logger

logger.debug(colored("Loading zrb.runner", attrs=["dark"]))
runner = Runner(env_prefix=ENV_PREFIX)



import click

from zrb.helper.typecheck import typechecked


@typechecked
def edit(editor: str, mark_comment: str, text: str, extension: str = "txt") -> str:
    result = click.edit(
        text="\n".join([mark_comment, text]), editor=editor, extension=f".{extension}"
    )
    if result is None:
        result = text
    return "\n".join(result.split(mark_comment)).strip()



from logging import Logger

from component.error import HTTPAPIException
from component.messagebus import Publisher
from component.rpc import Caller
from fastapi import Depends, FastAPI
from module.auth.component import Authorizer
from module.auth.integration import access_token_scheme
from module.auth.schema.permission import Permission, PermissionData, PermissionResult
from module.auth.schema.token import AccessTokenData
from opentelemetry import trace

tracer = trace.get_tracer(__name__)


def register_api(
    logger: Logger,
    app: FastAPI,
    authorizer: Authorizer,
    rpc_caller: Caller,
    publisher: Publisher,
):
    logger.info('🥪 Register API for "auth.permission"')

    @app.get("/api/v1/auth/permissions", response_model=PermissionResult)
    async def get_permissions(
        keyword: str = "",
        limit: int = 100,
        offset: int = 0,
        user_token_data: AccessTokenData = Depends(access_token_scheme),
    ):
        with tracer.start_as_current_span("authorizer.is_having_permission"):
            if not await authorizer.is_having_permission(
                user_token_data.user_id, "auth:permission:get"
            ):
                raise HTTPAPIException(403, "Unauthorized")
        try:
            with tracer.start_as_current_span("auth.rpc.auth_get_permission"):
                result_dict = await rpc_caller.call(
                    "auth_get_permission",
                    keyword=keyword,
                    criterion={},
                    limit=limit,
                    offset=offset,
                    user_token_data=user_token_data.model_dump(),
                )
                return PermissionResult(**result_dict)
        except Exception as e:
            raise HTTPAPIException(error=e)

    @app.get("/api/v1/auth/permissions/{id}", response_model=Permission)
    async def get_permission_by_id(
        id: str, user_token_data: AccessTokenData = Depends(access_token_scheme)
    ):
        with tracer.start_as_current_span("authorizer.is_having_permission"):
            if not await authorizer.is_having_permission(
                user_token_data.user_id, "auth:permission:get_by_id"
            ):
                raise HTTPAPIException(403, "Unauthorized")
        try:
            with tracer.start_as_current_span("auth.rpc.auth_get_permission_by_id"):
                result_dict = await rpc_caller.call(
                    "auth_get_permission_by_id",
                    id=id,
                    user_token_data=user_token_data.model_dump(),
                )
                return Permission(**result_dict)
        except Exception as e:
            raise HTTPAPIException(error=e)

    @app.post("/api/v1/auth/permissions", response_model=Permission)
    async def insert_permission(
        data: PermissionData,
        user_token_data: AccessTokenData = Depends(access_token_scheme),
    ):
        with tracer.start_as_current_span("authorizer.is_having_permission"):
            if not await authorizer.is_having_permission(
                user_token_data.user_id, "auth:permission:insert"
            ):
                raise HTTPAPIException(403, "Unauthorized")
        try:
            with tracer.start_as_current_span("auth.rpc.auth_insert_permission"):
                result_dict = await rpc_caller.call(
                    "auth_insert_permission",
                    data=data.model_dump(),
                    user_token_data=user_token_data.model_dump(),
                )
                return Permission(**result_dict)
        except Exception as e:
            raise HTTPAPIException(error=e)

    @app.put("/api/v1/auth/permissions/{id}", response_model=Permission)
    async def update_permission(
        id: str,
        data: PermissionData,
        user_token_data: AccessTokenData = Depends(access_token_scheme),
    ):
        with tracer.start_as_current_span("authorizer.is_having_permission"):
            if not await authorizer.is_having_permission(
                user_token_data.user_id, "auth:permission:update"
            ):
                raise HTTPAPIException(403, "Unauthorized")
        try:
            with tracer.start_as_current_span("auth.rpc.auth_update_permission"):
                result_dict = await rpc_caller.call(
                    "auth_update_permission",
                    id=id,
                    data=data.model_dump(),
                    user_token_data=user_token_data.model_dump(),
                )
                return Permission(**result_dict)
        except Exception as e:
            raise HTTPAPIException(error=e)

    @app.delete("/api/v1/auth/permissions/{id}", response_model=Permission)
    async def delete_permission(
        id: str, user_token_data: AccessTokenData = Depends(access_token_scheme)
    ):
        with tracer.start_as_current_span("authorizer.is_having_permission"):
            if not await authorizer.is_having_permission(
                user_token_data.user_id, "auth:permission:delete"
            ):
                raise HTTPAPIException(403, "Unauthorized")
        try:
            with tracer.start_as_current_span("auth.rpc.auth_delete_permission"):
                result_dict = await rpc_caller.call(
                    "auth_delete_permission",
                    id=id,
                    user_token_data=user_token_data.model_dump(),
                )
                return Permission(**result_dict)
        except Exception as e:
            raise HTTPAPIException(error=e)



from collections.abc import Callable, Iterable
from typing import Optional, TypeVar, Union

from zrb.helper.accessories.color import colored
from zrb.helper.log import logger
from zrb.helper.typecheck import typechecked
from zrb.task.any_task import AnyTask
from zrb.task.any_task_event_handler import (
    OnFailed,
    OnReady,
    OnRetry,
    OnSkipped,
    OnStarted,
    OnTriggered,
    OnWaiting,
)
from zrb.task.base_task.base_task import BaseTask
from zrb.task_env.env import Env
from zrb.task_env.env_file import EnvFile
from zrb.task_group.group import Group
from zrb.task_input.any_input import AnyInput

logger.debug(colored("Loading zrb.task.flow_task", attrs=["dark"]))

TFlowTask = TypeVar("TFlowTask", bound="FlowTask")


@typechecked
class FlowTask(BaseTask):
    def __init__(
        self,
        name: str,
        group: Optional[Group] = None,
        inputs: Iterable[AnyInput] = [],
        envs: Iterable[Env] = [],
        env_files: Iterable[EnvFile] = [],
        icon: Optional[str] = None,
        color: Optional[str] = None,
        description: str = "",
        upstreams: Iterable[AnyTask] = [],
        fallbacks: Iterable[AnyTask] = [],
        on_triggered: Optional[OnTriggered] = None,
        on_waiting: Optional[OnWaiting] = None,
        on_skipped: Optional[OnSkipped] = None,
        on_started: Optional[OnStarted] = None,
        on_ready: Optional[OnReady] = None,
        on_retry: Optional[OnRetry] = None,
        on_failed: Optional[OnFailed] = None,
        checkers: Iterable[AnyTask] = [],
        checking_interval: Union[float, int] = 0.05,
        retry: int = 2,
        retry_interval: Union[float, int] = 1,
        steps: list[Union[AnyTask, list[AnyTask]]] = [],
        should_execute: Union[bool, str, Callable[..., bool]] = True,
        return_upstream_result: bool = False,
    ):
        BaseTask.__init__(
            self,
            name=name,
            group=group,
            inputs=inputs,
            envs=envs,
            env_files=env_files,
            icon=icon,
            color=color,
            description=description,
            upstreams=self._create_flow_upstreams(
                steps=steps,
                upstreams=list(upstreams),
                inputs=list(inputs),
                envs=list(envs),
                env_files=list(env_files),
            ),
            fallbacks=fallbacks,
            on_triggered=on_triggered,
            on_waiting=on_waiting,
            on_skipped=on_skipped,
            on_started=on_started,
            on_ready=on_ready,
            on_retry=on_retry,
            on_failed=on_failed,
            checkers=checkers,
            checking_interval=checking_interval,
            retry=retry,
            retry_interval=retry_interval,
            should_execute=should_execute,
            return_upstream_result=return_upstream_result,
        )

    def copy(self) -> TFlowTask:
        return super().copy()

    def _create_flow_upstreams(
        self,
        steps: list[Union[AnyTask, list[AnyTask]]],
        upstreams: list[AnyTask],
        inputs: list[AnyInput],
        envs: list[Env],
        env_files: list[EnvFile],
    ) -> list[AnyTask]:
        flow_upstreams = upstreams
        for step in steps:
            tasks = [task.copy() for task in self._step_to_tasks(step)]
            new_upstreams = self._create_embeded_tasks(
                tasks=tasks,
                upstreams=flow_upstreams,
                inputs=inputs,
                envs=envs,
                env_files=env_files,
            )
            flow_upstreams = new_upstreams
        return flow_upstreams

    def _step_to_tasks(self, step: Union[AnyTask, list[AnyTask]]) -> list[AnyTask]:
        if isinstance(step, AnyTask):
            return [step]
        return step

    def _create_embeded_tasks(
        self,
        tasks: list[AnyTask],
        upstreams: list[AnyTask],
        inputs: list[AnyInput],
        envs: list[Env],
        env_files: list[EnvFile],
    ) -> list[AnyTask]:
        embeded_tasks: list[AnyTask] = []
        for embeded_task in tasks:
            embeded_task_upstreams = self._get_all_upstreams(tasks=[embeded_task])
            for embeded_task_upstream in embeded_task_upstreams:
                embeded_task_upstream.add_upstream(*upstreams)
            embeded_task.add_env(*envs)
            embeded_task.add_env_file(*env_files)
            embeded_task.add_input(*inputs)
            embeded_tasks.append(embeded_task)
        return embeded_tasks

    def _get_all_upstreams(self, tasks: list[AnyTask]):
        all_upstreams = []
        for task in tasks:
            upstreams = task._get_upstreams()
            if len(upstreams) == 0:
                all_upstreams.append(task)
                continue
            for upstream in upstreams:
                if len(upstream._get_upstreams()) == 0:
                    all_upstreams.append(upstream)
                    continue
                all_upstreams += self._get_all_upstreams([upstream])
        return all_upstreams



from collections.abc import Mapping
from logging import Logger
from typing import Any

from component.messagebus import Publisher
from component.repo import SearchFilter
from component.rpc import Caller, Server
from module.auth.schema.token import AccessTokenData
from module.snake_zrb_module_name.integration.model.snake_zrb_entity_name_model import (
    snake_zrb_entity_name_model,
)
from module.snake_zrb_module_name.schema.snake_zrb_entity_name import (
    PascalZrbEntityNameData,
)


def register_rpc(
    logger: Logger, rpc_server: Server, rpc_caller: Caller, publisher: Publisher
):
    logger.info(
        '🥪 Register RPC handlers for "snake_zrb_module_name.snake_zrb_entity_name"'
    )

    @rpc_server.register("snake_zrb_module_name_get_snake_zrb_entity_name")
    async def get(
        keyword: str,
        criterion: Mapping[str, Any],
        limit: int,
        offset: int,
        user_token_data: Mapping[str, Any],
    ) -> Mapping[str, Any]:
        result = await snake_zrb_entity_name_model.get(
            search_filter=SearchFilter(keyword=keyword, criterion=criterion),
            limit=limit,
            offset=offset,
        )
        return result.model_dump()

    @rpc_server.register("snake_zrb_module_name_get_snake_zrb_entity_name_by_id")
    async def get_by_id(
        id: str, user_token_data: Mapping[str, Any] = {}
    ) -> Mapping[str, Any]:
        row = await snake_zrb_entity_name_model.get_by_id(id)
        return row.model_dump()

    @rpc_server.register("snake_zrb_module_name_insert_snake_zrb_entity_name")
    async def insert(
        data: Mapping[str, Any], user_token_data: Mapping[str, Any]
    ) -> Mapping[str, Any]:
        user_token_data = AccessTokenData(**user_token_data)
        data["created_by"] = user_token_data.user_id
        data["updated_by"] = user_token_data.user_id
        row = await snake_zrb_entity_name_model.insert(
            data=PascalZrbEntityNameData(**data)
        )
        return row.model_dump()

    @rpc_server.register("snake_zrb_module_name_update_snake_zrb_entity_name")
    async def update(
        id: str, data: Mapping[str, Any], user_token_data: Mapping[str, Any]
    ) -> Mapping[str, Any]:
        user_token_data = AccessTokenData(**user_token_data)
        data["updated_by"] = user_token_data.user_id
        row = await snake_zrb_entity_name_model.update(
            id=id, data=PascalZrbEntityNameData(**data)
        )
        return row.model_dump()

    @rpc_server.register("snake_zrb_module_name_delete_snake_zrb_entity_name")
    async def delete(id: str, user_token_data: Mapping[str, Any]) -> Mapping[str, Any]:
        user_token_data = AccessTokenData(**user_token_data)
        row = await snake_zrb_entity_name_model.delete(id=id)
        return row.model_dump()



from ._group import snake_zrb_app_name_container_group
from .microservices import (
    snake_zrb_app_name_microservices_container_group,
    start_snake_zrb_app_name_microservices_container,
)
from .monolith import (
    snake_zrb_app_name_monolith_container_group,
    start_snake_zrb_app_name_monolith_container,
)
from .remove import remove_snake_zrb_app_name_container
from .stop import stop_snake_zrb_app_name_container
from .support import (
    snake_zrb_app_name_support_container_group,
    start_snake_zrb_app_name_support_container,
)

assert snake_zrb_app_name_container_group
assert snake_zrb_app_name_microservices_container_group
assert snake_zrb_app_name_monolith_container_group
assert snake_zrb_app_name_support_container_group
assert start_snake_zrb_app_name_microservices_container
assert start_snake_zrb_app_name_monolith_container
assert start_snake_zrb_app_name_support_container
assert remove_snake_zrb_app_name_container
assert stop_snake_zrb_app_name_container



from zrb.config.config import DEFAULT_SHELL
from zrb.task_input.str_input import StrInput

terminal_config_file_input = StrInput(
    name="config-file",
    shortcut="c",
    prompt="Config file",
    default="~/.zshrc" if DEFAULT_SHELL == "zsh" else "~/.bashrc",
)



import fnmatch
import glob
from collections.abc import Iterable

from zrb.helper.accessories.color import colored
from zrb.helper.log import logger
from zrb.helper.typecheck import typechecked

logger.debug(colored("Loading zrb.helper.file.match", attrs=["dark"]))


@typechecked
def get_file_names(glob_path: str, glob_ignored_paths: Iterable[str]) -> list[str]:
    matches = []
    for file in glob.glob(glob_path, recursive=True):
        should_ignore = any(
            fnmatch.fnmatch(file, ignored_path)
            or fnmatch.fnmatch(file, ignored_path + "**")
            for ignored_path in glob_ignored_paths
        )
        if not should_ignore:
            matches.append(file)
    return matches



import os
from collections.abc import Callable, Iterable, Mapping
from typing import Any, Optional, TypeVar, Union

from zrb.helper.accessories.color import colored
from zrb.helper.file.match import get_file_names
from zrb.helper.log import logger
from zrb.helper.typecheck import typechecked
from zrb.helper.typing import JinjaTemplate
from zrb.task.any_task import AnyTask
from zrb.task.any_task_event_handler import (
    OnFailed,
    OnReady,
    OnRetry,
    OnSkipped,
    OnStarted,
    OnTriggered,
    OnWaiting,
)
from zrb.task.watcher import Watcher
from zrb.task_env.env import Env
from zrb.task_env.env_file import EnvFile
from zrb.task_group.group import Group
from zrb.task_input.any_input import AnyInput

logger.debug(colored("Loading zrb.task.path_watcher", attrs=["dark"]))

TPathWatcher = TypeVar("TPathWatcher", bound="PathWatcher")


@typechecked
class PathWatcher(Watcher):
    """
    PathWatcher will wait for any changes specified on  path.

    Once the changes detected, PathWatcher will be completed
    and several xcom will be set:
    - <task-name>.file
    - <task-name>.new-file
    - <task-name>.modified-file
    - <task-name>.deleted-file
    """

    __init_times: Mapping[str, Mapping[str, float]] = {}

    def __init__(
        self,
        name: str = "watch-path",
        group: Optional[Group] = None,
        inputs: Iterable[AnyInput] = [],
        envs: Iterable[Env] = [],
        env_files: Iterable[EnvFile] = [],
        icon: Optional[str] = None,
        color: Optional[str] = None,
        description: str = "",
        upstreams: Iterable[AnyTask] = [],
        fallbacks: Iterable[AnyTask] = [],
        on_triggered: Optional[OnTriggered] = None,
        on_waiting: Optional[OnWaiting] = None,
        on_skipped: Optional[OnSkipped] = None,
        on_started: Optional[OnStarted] = None,
        on_ready: Optional[OnReady] = None,
        on_retry: Optional[OnRetry] = None,
        on_failed: Optional[OnFailed] = None,
        path: JinjaTemplate = "",
        ignored_path: Union[JinjaTemplate, Iterable[JinjaTemplate]] = [],
        checking_interval: Union[int, float] = 0.05,
        progress_interval: Union[int, float] = 30,
        watch_new_files: bool = True,
        watch_modified_files: bool = True,
        watch_deleted_files: bool = True,
        should_execute: Union[bool, JinjaTemplate, Callable[..., bool]] = True,
    ):
        Watcher.__init__(
            self,
            name=name,
            group=group,
            inputs=inputs,
            envs=envs,
            env_files=env_files,
            icon=icon,
            color=color,
            description=description,
            upstreams=upstreams,
            fallbacks=fallbacks,
            on_triggered=on_triggered,
            on_waiting=on_waiting,
            on_skipped=on_skipped,
            on_started=on_started,
            on_ready=on_ready,
            on_retry=on_retry,
            on_failed=on_failed,
            checking_interval=checking_interval,
            progress_interval=progress_interval,
            should_execute=should_execute,
        )
        self._path = path
        self._ignored_path = ignored_path
        self._watch_new_files = watch_new_files
        self._watch_modified_files = watch_modified_files
        self._watch_deleted_files = watch_deleted_files
        self._rendered_path: str = ""
        self._rendered_ignored_paths: list[str] = []

    def copy(self) -> TPathWatcher:
        return super().copy()

    def to_function(
        self,
        env_prefix: str = "",
        raise_error: bool = True,
        is_async: bool = False,
        show_done_info: bool = True,
        should_clear_xcom: bool = False,
        should_stop_looper: bool = False,
    ) -> Callable[..., bool]:
        return super().to_function(
            env_prefix=env_prefix,
            raise_error=raise_error,
            is_async=is_async,
            show_done_info=show_done_info,
            should_clear_xcom=should_clear_xcom,
            should_stop_looper=should_stop_looper,
        )

    async def run(self, *args: Any, **kwargs: Any) -> bool:
        self._rendered_path = self.render_str(self._path)
        self._rendered_ignored_paths = [
            ignored_path
            for ignored_path in self._get_rendered_ignored_paths()
            if ignored_path != ""
        ]
        identifier = self.get_identifier()
        if identifier not in self.__init_times:
            self.__init_times[identifier] = self._get_mod_times()
        return await super().run(*args, **kwargs)

    def _get_rendered_ignored_paths(self) -> list[str]:
        if isinstance(self._ignored_path, str):
            return [self.render_str(self._ignored_path)]
        return [self.render_str(ignored_path) for ignored_path in self._ignored_path]

    def create_loop_inspector(self) -> Callable[..., Optional[bool]]:
        def loop_inspect() -> bool:
            label = f"Watching {self._rendered_path}"
            identifier = self.get_identifier()
            try:
                init_times = self.__init_times[identifier]
                mod_times = self._get_mod_times()
            except Exception as e:
                self.show_progress(f"{label} Cannot inspect")
                raise e
            # watch changes
            if self._watch_new_files:
                new_files = mod_times.keys() - init_times.keys()
                for file in new_files:
                    self.print_out_dark(f"{label} [+] New file detected: {file}")
                    self.set_task_xcom("new-file", file)
                    self.set_task_xcom("file", file)
                    self.__init_times[identifier] = self._get_mod_times()
                    return True
            if self._watch_deleted_files:
                deleted_files = init_times.keys() - mod_times.keys()
                for file in deleted_files:
                    self.print_out_dark(f"{label} [-] File deleted: {file}")
                    self.set_task_xcom("deleted-file", file)
                    self.set_task_xcom("file", file)
                    self.__init_times[identifier] = self._get_mod_times()
                    return True
            if self._watch_modified_files:
                modified_files = {
                    file
                    for file, mod_time in mod_times.items()
                    if file in mod_times and init_times[file] != mod_time
                }
                for file in modified_files:
                    self.print_out_dark(f"{label} [/] File modified: {file}")
                    self.set_task_xcom("modified-file", file)
                    self.set_task_xcom("file", file)
                    self.__init_times[identifier] = self._get_mod_times()
                    return True
            self.show_progress(f"{label} (Nothing changed)")
            return False

        return loop_inspect

    def _get_mod_times(self) -> Mapping[str, float]:
        matches = get_file_names(
            glob_path=self._rendered_path,
            glob_ignored_paths=self._rendered_ignored_paths,
        )
        mod_times: Mapping[str, float] = {}
        for file_name in matches:
            try:
                mod_time = os.stat(file_name).st_mtime
                mod_times[file_name] = mod_time
            except Exception as e:
                self.print_err(e)
        return mod_times



import os

from zrb import CmdTask, Env, EnvFile, StrInput, python_task, runner

from ._constant import APP_TEMPLATE_ENV_FILE_NAME, RESOURCE_DIR
from ._group import snake_zrb_app_name_group
from .backend import prepare_snake_zrb_app_name_backend
from .frontend import build_snake_zrb_app_name_frontend_once

_CURRENT_DIR = os.path.dirname(__file__)


@python_task(
    icon="🧪",
    name="remove-test-db",
    group=snake_zrb_app_name_group,
    runner=runner,
)
def remove_snake_zrb_app_name_test_db(*args, **kwargs):
    test_db_file_path = os.path.join(RESOURCE_DIR, "test.db")
    if os.path.isfile(test_db_file_path):
        os.remove(test_db_file_path)


test_snake_zrb_app_name = CmdTask(
    icon="🚤",
    name="test",
    group=snake_zrb_app_name_group,
    inputs=[
        StrInput(
            name="kebab-zrb-app-name-test",
            description="Specific test case (i.e., test/file.py::test_name)",
            prompt="Test (i.e., test/file.py::test_name)",
            default="",
        )
    ],
    upstreams=[
        build_snake_zrb_app_name_frontend_once,
        prepare_snake_zrb_app_name_backend,
        remove_snake_zrb_app_name_test_db,
    ],
    cwd=RESOURCE_DIR,
    env_files=[EnvFile(path=APP_TEMPLATE_ENV_FILE_NAME, prefix="TEST_ZRB_ENV_PREFIX")],
    envs=[
        Env(name="APP_BROKER_TYPE", os_name="TEST_APP_BROKER_TYPE", default="mock"),
        Env(
            name="APP_DB_CONNECTION",
            os_name="TEST_APP_DB_CONNECTION",
            default="sqlite:///test.db",
        ),
        Env(name="APP_AUTH_ADMIN_ACTIVE", os_name="", default="true"),
        Env(name="APP", os_name="APP_ENABLE_OTEL", default="false"),
    ],
    cmd_path=os.path.join(_CURRENT_DIR, "test.sh"),
    should_print_cmd_result=False,
    retry=0,
)
runner.register(test_snake_zrb_app_name)



from collections.abc import Mapping
from typing import Any

from ruamel.yaml import YAML, CommentedMap

from zrb.helper.accessories.color import colored
from zrb.helper.log import logger
from zrb.helper.typecheck import typechecked

logger.debug(colored("Loading zrb.helper.docker_compose.file", attrs=["dark"]))


@typechecked
def read_compose_file(file_name: str) -> Any:
    yaml = YAML()
    with open(file_name, "r") as f:
        data = yaml.load(f)
    return data


@typechecked
def write_compose_file(file_name: str, data: Any):
    yaml = YAML()
    with open(file_name, "w") as f:
        yaml.dump(data, f)


@typechecked
def add_services(file_name: str, new_services: Mapping[str, str]):
    data = read_compose_file(file_name)
    data = CommentedMap(data)
    data["services"].update(CommentedMap(new_services))
    write_compose_file(file_name, data)



from zrb.helper.accessories.color import colored
from zrb.helper.log import logger
from zrb.helper.typecheck import typechecked

logger.debug(colored("Loading zrb.helper.string.modification", attrs=["dark"]))


@typechecked
def double_quote(text: str) -> str:
    return '"' + text.replace('"', '\\"') + '"'



import importlib.metadata as metadata
import os

from zrb.helper.string.untyped_conversion import (
    untyped_to_boolean,
    untyped_to_logging_level,
)


def _get_version() -> str:
    return metadata.version("zrb")


def _get_current_shell() -> str:
    current_shell = os.getenv("SHELL", "")
    if current_shell.endswith("zsh"):
        return "zsh"
    return "bash"


def _get_valid_container_backend(container_backend: str) -> str:
    if container_backend.lower().strip() == "podman":
        return "podman"
    return "docker"


def _get_default_tmp_dir() -> str:
    if os.path.isdir("/tmp"):
        return "/tmp"
    return os.path.expanduser(os.path.join("~", ".tmp"))


TMP_DIR = os.getenv("ZRB_TMP_DIR", _get_default_tmp_dir())
DEFAULT_SHELL = os.getenv("ZRB_SHELL", _get_current_shell())
DEFAULT_EDITOR = os.getenv("ZRB_EDITOR", "nano")
INIT_MODULES_STR = os.getenv("ZRB_INIT_MODULES", "")
INIT_MODULES = (
    [module.strip() for module in INIT_MODULES_STR.split(":") if module.strip() != ""]
    if INIT_MODULES_STR != ""
    else []
)
INIT_SCRIPTS_STR = os.getenv("ZRB_INIT_SCRIPTS", "")
INIT_SCRIPTS = (
    [script.strip() for script in INIT_SCRIPTS_STR.split(":") if script.strip() != ""]
    if INIT_SCRIPTS_STR != ""
    else []
)
LOGGING_LEVEL = untyped_to_logging_level(os.getenv("ZRB_LOGGING_LEVEL", "WARNING"))
SHOULD_LOAD_BUILTIN = untyped_to_boolean(os.getenv("ZRB_SHOULD_LOAD_BUILTIN", "1"))
ENV_PREFIX = os.getenv("ZRB_ENV", "")
SHOW_ADVERTISEMENT = untyped_to_boolean(os.getenv("ZRB_SHOW_ADVERTISEMENT", "1"))
SHOW_PROMPT = untyped_to_boolean(os.getenv("ZRB_SHOW_PROMPT", "1"))
SHOW_TIME = untyped_to_boolean(os.getenv("ZRB_SHOW_TIME", "1"))
VERSION = _get_version()
CONTAINER_BACKEND = _get_valid_container_backend(
    os.getenv("ZRB_CONTAINER_BACKEND", "docker")
)
ENABLE_TYPE_CHECKING = untyped_to_boolean(os.getenv("ZRB_ENABLE_TYPE_CHECKING", "1"))



from zrb import Group

from .._group import snake_zrb_app_name_container_group

snake_zrb_app_name_monolith_container_group = Group(
    name="monolith",
    parent=snake_zrb_app_name_container_group,
    description="Manage human readable zrb app name monolith containers",
)



import pytest
from asgi_lifespan import LifespanManager
from httpx import AsyncClient
from src.main import app


@pytest.fixture()
async def test_client_generator(scope="module"):
    async with LifespanManager(app):
        async with AsyncClient(app=app, base_url="http://localhost") as ac:
            yield ac



import logging

LOGGING_LEVEL_MAP = {
    "critical": logging.CRITICAL,
    "fatal": logging.FATAL,
    "error": logging.ERROR,
    "warning": logging.WARNING,
    "warn": logging.WARN,
    "info": logging.INFO,
    "debug": logging.DEBUG,
    "notset": logging.NOTSET,
}


def str_to_boolean(value: str):
    if value.lower() in ("0", "false", "no", "n"):
        return False
    if value.lower() in ("1", "true", "yes", "y"):
        return True
    raise Exception(f'Cannot convert to boolean: "{value}"')


def str_to_logging_level(logging_level_str: str) -> int:
    lower_logging_level_str = logging_level_str.lower()
    if lower_logging_level_str in LOGGING_LEVEL_MAP:
        return LOGGING_LEVEL_MAP[lower_logging_level_str]
    return logging.WARNING



from collections.abc import Callable, Iterable
from http.client import HTTPConnection, HTTPSConnection
from typing import Any, Optional, TypeVar, Union

from zrb.helper.accessories.color import colored
from zrb.helper.log import logger
from zrb.helper.typecheck import typechecked
from zrb.helper.typing import JinjaTemplate
from zrb.task.any_task import AnyTask
from zrb.task.any_task_event_handler import (
    OnFailed,
    OnReady,
    OnRetry,
    OnSkipped,
    OnStarted,
    OnTriggered,
    OnWaiting,
)
from zrb.task.checker import Checker
from zrb.task_env.env import Env
from zrb.task_env.env_file import EnvFile
from zrb.task_group.group import Group
from zrb.task_input.any_input import AnyInput

logger.debug(colored("Loading zrb.task.http_checker", attrs=["dark"]))

THTTPChecker = TypeVar("THTTPChecker", bound="HTTPChecker")


@typechecked
class HttpConnectionConfig:
    def __init__(
        self,
        is_https: bool,
        method: str,
        host: str,
        port: int,
        url: str,
        timeout: int,
    ):
        self.is_https = is_https
        self.method = method
        self.host = host
        self.port = port
        self.url = "/" + url if not url.startswith("/") else url
        self.timeout = timeout
        self.protocol = "https" if self.is_https else "http"
        full_url = f"{self.protocol}://{self.host}:{self.port}{self.url}"
        self.label = f"{self.method} {full_url}"

    def get_connection(self) -> Union[HTTPConnection, HTTPSConnection]:
        if self.is_https:
            return HTTPSConnection(host=self.host, port=self.port, timeout=self.timeout)
        return HTTPConnection(host=self.host, port=self.port, timeout=self.timeout)


@typechecked
class HTTPChecker(Checker):
    def __init__(
        self,
        name: str = "check-http",
        group: Optional[Group] = None,
        inputs: Iterable[AnyInput] = [],
        envs: Iterable[Env] = [],
        env_files: Iterable[EnvFile] = [],
        icon: Optional[str] = None,
        color: Optional[str] = None,
        description: str = "",
        host: JinjaTemplate = "localhost",
        port: Union[int, JinjaTemplate] = 80,
        timeout: Union[int, JinjaTemplate] = 5,
        method: str = "HEAD",
        url: str = "/",
        is_https: Union[bool, str] = False,
        upstreams: Iterable[AnyTask] = [],
        fallbacks: Iterable[AnyTask] = [],
        on_triggered: Optional[OnTriggered] = None,
        on_waiting: Optional[OnWaiting] = None,
        on_skipped: Optional[OnSkipped] = None,
        on_started: Optional[OnStarted] = None,
        on_ready: Optional[OnReady] = None,
        on_retry: Optional[OnRetry] = None,
        on_failed: Optional[OnFailed] = None,
        checking_interval: Union[int, float] = 0.05,
        progress_interval: Union[int, float] = 5,
        expected_result: bool = True,
        should_execute: Union[bool, JinjaTemplate, Callable[..., bool]] = True,
    ):
        Checker.__init__(
            self,
            name=name,
            group=group,
            inputs=inputs,
            envs=envs,
            env_files=env_files,
            icon=icon,
            color=color,
            description=description,
            upstreams=upstreams,
            fallbacks=fallbacks,
            on_triggered=on_triggered,
            on_waiting=on_waiting,
            on_skipped=on_skipped,
            on_started=on_started,
            on_ready=on_ready,
            on_retry=on_retry,
            on_failed=on_failed,
            checking_interval=checking_interval,
            progress_interval=progress_interval,
            expected_result=expected_result,
            should_execute=should_execute,
        )
        self._host = host
        self._port = port
        self._timeout = timeout
        self._method = method
        self._url = url
        self._is_https = is_https
        self._config: Optional[HttpConnectionConfig] = None

    def copy(self) -> THTTPChecker:
        return super().copy()

    def to_function(
        self,
        env_prefix: str = "",
        raise_error: bool = True,
        is_async: bool = False,
        show_done_info: bool = True,
        should_clear_xcom: bool = False,
        should_stop_looper: bool = False,
    ) -> Callable[..., bool]:
        return super().to_function(
            env_prefix=env_prefix,
            raise_error=raise_error,
            is_async=is_async,
            show_done_info=show_done_info,
            should_clear_xcom=should_clear_xcom,
            should_stop_looper=should_stop_looper,
        )

    async def run(self, *args: Any, **kwargs: Any) -> bool:
        self._config = HttpConnectionConfig(
            is_https=self.render_bool(self._is_https),
            method=self.render_str(self._method),
            host=self.render_str(self._host),
            port=self.render_int(self._port),
            url=self.render_str(self._url),
            timeout=self.render_int(self._timeout),
        )
        return await super().run(*args, **kwargs)

    async def inspect(self, *args: Any, **kwargs: Any) -> bool:
        conn = self._config.get_connection()
        try:
            conn.request(self._config.method, self._config.url)
            res = conn.getresponse()
            if res.status < 300:
                self.log_info("Connection success")
                self.print_out(f"{self._config.label} {res.status} (OK)")
                return True
            self.show_progress(f"{self._config.label} {res.status} (Not OK)")
        except Exception:
            self.show_progress(f"{self._config.label} Connection error")
        finally:
            conn.close()
        return False



from typing import Any

from zrb.helper.accessories.color import colored
from zrb.helper.log import logger
from zrb.helper.typecheck import typechecked

logger.debug(colored("Loading zrb.helper.string.jinja", attrs=["dark"]))


@typechecked
def is_probably_jinja(value: Any) -> bool:
    if not isinstance(value, str):
        return False
    if "{{" in value and "}}" in value:
        return True
    if "{%" in value and "%}" in value:
        return True
    return False



import importlib
from functools import lru_cache
from typing import Any

from zrb.helper.accessories.color import colored
from zrb.helper.log import logger
from zrb.helper.typecheck import typechecked

logger.debug(colored("Loading zrb.helper.loader.load_module", attrs=["dark"]))


@lru_cache
@typechecked
def load_module(module_name: str) -> Any:
    logger.info(colored(f"Importing module: {module_name}", attrs=["dark"]))
    module = importlib.import_module(module_name)
    logger.info(colored(f"Module imported: {module_name}", attrs=["dark"]))
    return module



import os

from zrb import CmdTask, runner
from ..._project import build_project
from .._group import snake_zrb_package_name_group
from ..prepare_venv import prepare_snake_zrb_package_name_venv

_CURRENT_DIR = os.path.dirname(__file__)
_PKG_AUTOMATE_DIR = os.path.dirname(_CURRENT_DIR)
_PROJECT_DIR = os.path.dirname(os.path.dirname(_PKG_AUTOMATE_DIR))
_PKG_RESOURCE_DIR = os.path.join(_PROJECT_DIR, "src", "kebab-zrb-package-name")

build_snake_zrb_package_name = CmdTask(
    name="build",
    description="Build human readable zrb package name",
    group=snake_zrb_package_name_group,
    upstreams=[prepare_snake_zrb_package_name_venv],
    cwd=_PKG_RESOURCE_DIR,
    cmd_path=[
        os.path.join(_PKG_AUTOMATE_DIR, "activate-venv.sh"),
        os.path.join(_CURRENT_DIR, "build.sh"),
    ],
    should_print_cmd_result=False,
)

build_snake_zrb_package_name >> build_project

runner.register(build_snake_zrb_package_name)



import json
import os
from collections.abc import Mapping

from zrb.builtin.monorepo._config import (
    MONOREPO_CONFIG,
    MONOREPO_CONFIG_FILE,
    PROJECT_DIR,
)
from zrb.builtin.monorepo._group import monorepo_group
from zrb.runner import runner
from zrb.task.any_task import AnyTask
from zrb.task.decorator import python_task
from zrb.task_input.str_input import StrInput


@python_task(
    name="add",
    group=monorepo_group,
    inputs=[
        StrInput(name="alias", shortcut="a", prompt="Repo Alias", prompt_required=True),
        StrInput(name="folder", shortcut="d", prompt="Directory", prompt_required=True),
        StrInput(name="origin", shortcut="o", prompt="Repo URL", prompt_required=True),
        StrInput(
            name="branch",
            shortcut="b",
            prompt="Repo Branch",
            prompt_required=True,
            default="main",
        ),
    ],
    description="Add repo to monorepo",
    runner=runner,
)
def add_to_monorepo(*args, **kwargs):
    task: AnyTask = kwargs.get("_task")
    input_map: Mapping[str, str] = task.get_input_map()
    abs_folder = os.path.join(PROJECT_DIR, input_map.get("folder", ""))
    if os.path.isdir(abs_folder):
        raise ValueError(f"Directory exists: {abs_folder}")
    config = dict(MONOREPO_CONFIG)
    config[input_map.get("alias", "")] = {
        "folder": input_map.get("folder", ""),
        "branch": input_map.get("branch", ""),
        "origin": input_map.get("origin", ""),
    }
    with open(MONOREPO_CONFIG_FILE, "w") as file:
        json.dump(config, file, indent=2)



import os
from typing import Any

from zrb.builtin.project._input import project_dir_input
from zrb.helper.codemod.add_assert_resource import add_assert_resource
from zrb.helper.codemod.add_import_module import add_import_module
from zrb.helper.file.copy_tree import copy_tree
from zrb.helper.file.text import read_text_file_async, write_text_file_async
from zrb.task.decorator import python_task
from zrb.task.task import Task

_CURRENT_DIR = os.path.dirname(__file__)


@python_task(
    name="add-project-tasks",
    inputs=[project_dir_input],
)
async def add_project_tasks(*args: Any, **kwargs: Any):
    task: Task = kwargs.get("_task")
    project_dir = kwargs.get("project_dir", ".")
    task.print_out("Create project tasks if not exist")
    if os.path.exists(os.path.join(project_dir, "_automate", "_project")):
        return
    await copy_tree(
        src=os.path.join(_CURRENT_DIR, "template"),
        dst=project_dir,
        excludes=["*/__pycache__"],
    )
    zrb_init_path = os.path.join(project_dir, "zrb_init.py")
    code = await read_text_file_async(zrb_init_path)
    import_alias = "_automate_project"
    code = add_import_module(
        code=code, module_path="_automate._project", alias="_automate_project"
    )
    code = add_assert_resource(code, import_alias)
    await write_text_file_async(zrb_init_path, code)



from collections.abc import Callable, Iterable, Mapping
from typing import Any, Optional, Type, TypeVar, Union

from zrb.helper.accessories.untyped_color import untyped_colored as colored
from zrb.helper.log import logger

logger.debug(colored("Loading zrb.helper.typing", attrs=["dark"]))

Tuple = tuple
List = list

JinjaTemplate = str

assert Any
assert Iterable
assert Callable
assert List
assert Mapping
assert Optional
assert TypeVar
assert Type
assert Tuple
assert Union



import os
import sys

from ._group import snake_zrb_package_name_group
from .build import build_snake_zrb_package_name
from .prepare_venv.prepare_venv import prepare_snake_zrb_package_name_venv
from .publish.publish import publish_snake_zrb_package_name

_CURRENT_DIR = os.path.dirname(__file__)
_AUTOMATE_DIR = os.path.dirname(_CURRENT_DIR)
_PROJECT_DIR = os.path.dirname(_AUTOMATE_DIR)
_PLUGIN_SRC_DIR = os.path.join(_PROJECT_DIR, "src", "kebab-zrb-package-name", "src")
if os.path.isdir(_PLUGIN_SRC_DIR):
    sys.path.append(_PLUGIN_SRC_DIR)
    import snake_zrb_package_name

    assert snake_zrb_package_name

assert snake_zrb_package_name_group
assert build_snake_zrb_package_name
assert prepare_snake_zrb_package_name_venv
assert publish_snake_zrb_package_name



from ._group import snake_zrb_app_name_container_group
from .remove import remove_snake_zrb_app_name_container
from .start import start_snake_zrb_app_name_container
from .stop import stop_snake_zrb_app_name_container

assert snake_zrb_app_name_container_group
assert remove_snake_zrb_app_name_container
assert start_snake_zrb_app_name_container
assert stop_snake_zrb_app_name_container



from zrb.builtin.project._group import project_group
from zrb.builtin.project.add import (
    add_cmd_task,
    add_docker_compose_task,
    add_fastapp_application,
    add_fastapp_crud,
    add_fastapp_field,
    add_fastapp_module,
    add_plugin,
    add_project_tasks,
    add_python_app,
    add_python_task,
    project_add_app_group,
    project_add_fastapp_group,
    project_add_group,
    project_add_task_group,
)
from zrb.builtin.project.create import create_project

assert project_group
assert project_add_group
assert project_add_app_group
assert project_add_fastapp_group
assert project_add_task_group
assert create_project
assert add_plugin
assert add_project_tasks
assert add_cmd_task
assert add_docker_compose_task
assert add_fastapp_application
assert add_fastapp_crud
assert add_fastapp_module
assert add_fastapp_field
assert add_python_task
assert add_python_app



from zrb.runner import runner
from zrb.task.cmd_task import CmdTask

update_zrb = CmdTask(
    name="update",
    description="Update zrb",
    cmd="pip install zrb -U",
    checking_interval=3,
)
runner.register(update_zrb)



from zrb import BoolInput

enable_monitoring_input = BoolInput(
    name="enable-kebab-zrb-app-name-monitoring",
    description='Enable "kebab-zrb-app-name" monitoring',
    prompt='Enable "kebab-zrb-app-name" monitoring?',
    default=False,
)



import os

from zrb.helper.util import coalesce, to_kebab_case, to_snake_case
from zrb.task_input.int_input import IntInput
from zrb.task_input.str_input import StrInput

SYSTEM_USER = os.getenv("USER", "incognito")

app_name_input = StrInput(
    name="app-name",
    shortcut="a",
    description="App name",
    prompt="App name",
    default="app",
)

app_author_name_input = StrInput(
    name="app-author-name",
    prompt="App author name",
    description="App author name",
    default=SYSTEM_USER,
)

app_author_email_input = StrInput(
    name="app-author-email",
    prompt="App author email",
    description="App author email",
    default=f"{SYSTEM_USER}@gmail.com",
)

app_description_input = StrInput(
    name="app-description",
    description="App description",
    prompt="App description",
    default="Just another app",
)

app_image_default_namespace = os.getenv(
    "PROJECT_IMAGE_DEFAULT_NAMESPACE", "docker.io/" + SYSTEM_USER
)

app_image_name_input = StrInput(
    name="app-image-name",
    description="App image name",
    prompt="App image name",
    default=lambda m: "/".join(
        [app_image_default_namespace, to_kebab_case(m.get("app_name"))]
    ),
)

http_port_input = IntInput(
    name="http-port",
    shortcut="p",
    description="HTTP port",
    prompt="HTTP port",
    default=8080,
)

env_prefix_input = StrInput(
    name="env-prefix",
    description="OS environment prefix",
    prompt="OS environment prefix",
    default=lambda m: to_snake_case(
        coalesce(m.get("app_name"), m.get("task_name"), "MY")
    ).upper(),
)



import asyncio

from config import APP_MAX_NOT_READY
from integration.log import logger


class AppState:
    __instance = None

    def __new__(cls):
        if cls.__instance is None:
            cls.__instance = super().__new__(cls)
        return cls.__instance

    def __init__(self):
        self.readiness: bool = False
        self.liveness: bool = False

    def set_liveness(self, value: bool):
        self.liveness = value

    def set_readiness(self, value: bool):
        self.readiness = value

    def get_liveness(self) -> bool:
        return self.liveness

    def get_readiness(self) -> bool:
        return self.readiness


app_state = AppState()


async def set_not_ready_on_error(exception: Exception):
    logger.critical(exception)
    app_state.set_readiness(False)
    await asyncio.sleep(APP_MAX_NOT_READY)
    app_state.set_liveness(False)



from config import (
    APP_AUTH_ADMIN_EMAIL,
    APP_AUTH_ADMIN_PASSWORD,
    APP_AUTH_ADMIN_PHONE,
    APP_AUTH_ADMIN_USER_ID,
    APP_AUTH_ADMIN_USERNAME,
    APP_AUTH_GUEST_EMAIL,
    APP_AUTH_GUEST_PHONE,
    APP_AUTH_GUEST_USER_ID,
    APP_AUTH_GUEST_USERNAME,
)
from module.auth.schema.user import User

admin_user_password = APP_AUTH_ADMIN_PASSWORD

admin_user = User(
    id=APP_AUTH_ADMIN_USER_ID,
    username=APP_AUTH_ADMIN_USERNAME,
    email=APP_AUTH_ADMIN_EMAIL,
    phone=APP_AUTH_ADMIN_PHONE,
    groups=[],
    permissions=[],
    description="System administrator",
)

guest_user = User(
    id=APP_AUTH_GUEST_USER_ID,
    username=APP_AUTH_GUEST_USERNAME,
    email=APP_AUTH_GUEST_EMAIL,
    phone=APP_AUTH_GUEST_PHONE,
    groups=[],
    permissions=[],
    description="Visitor",
)



from zrb.builtin.devtool._group import devtool_group
from zrb.builtin.devtool.install import (
    devtool_install_group,
    install_aws,
    install_docker,
    install_gcloud,
    install_gvm,
    install_helix,
    install_helm,
    install_kubectl,
    install_nvm,
    install_pulumi,
    install_pyenv,
    install_sdkman,
    install_selenium,
    install_terraform,
    install_tmux,
    install_zsh,
)

assert devtool_group
assert devtool_install_group
assert install_aws
assert install_docker
assert install_gcloud
assert install_gvm
assert install_helix
assert install_helm
assert install_kubectl
assert install_nvm
assert install_pulumi
assert install_pyenv
assert install_sdkman
assert install_selenium
assert install_terraform
assert install_tmux
assert install_zsh



from zrb.builtin.project.add.fastapp._group import project_add_fastapp_group
from zrb.builtin.project.add.fastapp.app import add_fastapp_application
from zrb.builtin.project.add.fastapp.crud import add_fastapp_crud
from zrb.builtin.project.add.fastapp.field import add_fastapp_field
from zrb.builtin.project.add.fastapp.module import add_fastapp_module

assert project_add_fastapp_group
assert add_fastapp_application
assert add_fastapp_module
assert add_fastapp_crud
assert add_fastapp_field



import json
import os
from collections.abc import Mapping

PROJECT_DIR = os.getenv("ZRB_PROJECT_DIR", ".")
MONOREPO_CONFIG_FILE = os.path.join(PROJECT_DIR, "monorepo.zrb.json")
MONOREPO_CONFIG: Mapping[str, Mapping[str, str]] = {}

if os.path.isfile(MONOREPO_CONFIG_FILE):
    with open(MONOREPO_CONFIG_FILE, "r") as file:
        MONOREPO_CONFIG = json.load(file)



from zrb import HTTPChecker, PortChecker

rabbitmq_management_checker = HTTPChecker(
    name="check-rabbitmq-management",
    port='{{env.get("RABBITMQ_MANAGEMENT_HOST_PORT", "15672")}}',
    is_https="{{input.snake_zrb_app_name_https}}",
    should_execute='{{env.get("APP_BROKER_TYPE", "rabbitmq") == "rabbitmq"}}',
)

rabbitmq_checker = PortChecker(
    name="check-rabbitmq",
    port='{{env.get("RABBITMQ_HOST_PORT", "5672")}}',
    should_execute='{{env.get("APP_BROKER_TYPE", "rabbitmq") == "rabbitmq"}}',
)

redpanda_console_checker = HTTPChecker(
    name="check-redpanda-console",
    method="GET",
    port='{{env.get("REDPANDA_CONSOLE_HOST_PORT", "9000")}}',
    is_https="{{input.snake_zrb_app_name_https}}",
    should_execute='{{env.get("APP_BROKER_TYPE", "rabbitmq") == "kafka"}}',
)

kafka_plaintext_checker = PortChecker(
    name="check-kafka-plaintext",
    port='{{env.get("KAFKA_INTERNAL_HOST_PORT", "29092")}}',
    should_execute='{{env.get("APP_BROKER_TYPE", "rabbitmq") == "kafka"}}',
)

kafka_outside_checker = PortChecker(
    name="check-kafka-outside",
    port='{{env.get("KAFKA_EXTERNAL_HOST_PORT", "9092")}}',
    should_execute='{{env.get("APP_BROKER_TYPE", "rabbitmq") == "kafka"}}',
)

pandaproxy_plaintext_checker = PortChecker(
    name="check-pandaproxy-plaintext",
    port='{{env.get("PANDAPROXY_INTERNAL_HOST_PORT", "29092")}}',
    should_execute='{{env.get("APP_BROKER_TYPE", "rabbitmq") == "kafka"}}',
)

pandaproxy_outside_checker = PortChecker(
    name="check-pandaproxy-outside",
    port='{{env.get("PANDAPROXY_EXTERNAL_HOST_PORT", "9092")}}',
    should_execute='{{env.get("APP_BROKER_TYPE", "rabbitmq") == "kafka"}}',
)

app_container_checker = HTTPChecker(
    name="check-kebab-zrb-app-name-container",
    host="{{input.snake_zrb_app_name_host}}",
    url="/readiness",
    port='{{env.get("HOST_PORT", "zrbAppHttpPort")}}',
    is_https="{{input.snake_zrb_app_name_https}}",
    should_execute="{{ input.local_snake_zrb_app_name}}",
)

app_local_checker = HTTPChecker(
    name="check-kebab-zrb-app-name",
    host="{{input.snake_zrb_app_name_host}}",
    url="/readiness",
    port='{{env.get("APP_PORT", "zrbAppHttpPort")}}',
    is_https="{{input.snake_zrb_app_name_https}}",
    should_execute="{{ input.local_snake_zrb_app_name}}",
)



from zrb import Env, EnvFile

from ._constant import (
    APP_TEMPLATE_ENV_FILE_NAME,
    DEPLOYMENT_DIR,
    DEPLOYMENT_TEMPLATE_ENV_FILE_NAME,
)

deployment_app_env_file = EnvFile(
    path=APP_TEMPLATE_ENV_FILE_NAME, prefix="DEPLOYMENT_APP_ZRB_ENV_PREFIX"
)

deployment_config_env_file = EnvFile(
    path=DEPLOYMENT_TEMPLATE_ENV_FILE_NAME, prefix="DEPLOYMENT_CONFIG_ZRB_ENV_PREFIX"
)

pulumi_backend_url_env = Env(
    name="PULUMI_BACKEND_URL",
    os_name="PULUMI_ZRB_ENV_PREFIX_BACKEND_URL",
    default=f"file://{DEPLOYMENT_DIR}/state",
)

pulumi_config_passphrase_env = Env(
    name="PULUMI_CONFIG_PASSPHRASE",
    os_name="PULUMI_ZRB_ENV_PREFIX_CONFIG_PASSPHRASE",
    default="secret",
)

deployment_replica_env = Env(
    name="REPLICA",
    os_name="DEPLOYMENT_ZRB_ENV_PREFIX",
    default="{{input.snake_zrb_app_name_replica}}",
)



from zrb.task.cmd_task import CmdTask
from zrb.task.notifier import Notifier
from zrb.task.task import Task
from zrb.task_input.str_input import StrInput


def create_recurring_action(
    notif_title: str,
    default_message: str = "👋",
) -> Task:
    # define inputs
    message_input = StrInput(
        name="message",
        default=default_message,
        prompt="Message to be shown",
    )
    command_input = StrInput(
        name="command",
        default="",
        prompt="Command to be executed",
    )
    run_command = CmdTask(
        name="run-command",
        icon="⚙️",
        color="blue",
        inputs=[command_input],
        should_execute='{{ input.command != "" }}',
        cmd="{{ input.command }}",
    )
    notify = Notifier(
        name="notify",
        icon="📢",
        color="green",
        inputs=[message_input],
        title=notif_title,
        message="{{ input.message }}",
        should_execute='{{ input.message != "" }}',
    )
    # return aggregator task
    return Task(
        name="recurring-action",
        inputs=[message_input, command_input],
        upstreams=[run_command, notify],
        retry=0,
    )



from zrb import StrInput

image_input = StrInput(
    name="kebab-zrb-app-name-image",
    description='Image name of "kebab-zrb-app-name"',
    prompt='Image name of "kebab-zrb-app-name"',
    default="zrb-app-image-name:latest",
)



from .build import build_project
from .container import (
    project_container_group,
    remove_project_containers,
    start_project_containers,
    stop_project_containers,
)
from .deploy import deploy_project
from .destroy import destroy_project
from .get_env import get_project_env
from .image import build_project_images, project_image_group, push_project_images
from .publish import publish_project
from .start import start_project

assert project_container_group
assert project_image_group
assert remove_project_containers
assert start_project_containers
assert stop_project_containers
assert build_project_images
assert push_project_images
assert build_project
assert deploy_project
assert destroy_project
assert get_project_env
assert publish_project
assert start_project



from component.messagebus.messagebus import Publisher
from module.auth.entity.permission.repo import PermissionRepo
from module.auth.schema.permission import Permission, PermissionData, PermissionResult
from module.log.component.historical_repo_model import HistoricalRepoModel


class PermissionModel(
    HistoricalRepoModel[Permission, PermissionData, PermissionResult]
):
    schema_result_cls = PermissionResult
    log_entity_name = "permission"

    def __init__(self, repo: PermissionRepo, publisher: Publisher):
        super().__init__(repo, publisher)

    async def ensure_permission(self, data: PermissionData):
        try:
            await self.repo.get_by_name(data.name)
        except Exception as e:
            error_message = f"{e}"
            if error_message.lower().startswith("not found"):
                await self.repo.insert(data)
                return
            raise e



from ._group import snake_zrb_app_name_group
from .container import (
    remove_snake_zrb_app_name_container,
    snake_zrb_app_name_container_group,
    start_snake_zrb_app_name_container,
    stop_snake_zrb_app_name_container,
)
from .deploy import deploy_snake_zrb_app_name
from .destroy import destroy_snake_zrb_app_name
from .image import (
    build_snake_zrb_app_name_image,
    push_snake_zrb_app_name_image,
    snake_zrb_app_name_image_group,
)
from .start import start_snake_zrb_app_name

assert snake_zrb_app_name_group
assert snake_zrb_app_name_image_group
assert snake_zrb_app_name_container_group
assert start_snake_zrb_app_name
assert deploy_snake_zrb_app_name
assert destroy_snake_zrb_app_name
assert remove_snake_zrb_app_name_container
assert stop_snake_zrb_app_name_container
assert start_snake_zrb_app_name_container
assert build_snake_zrb_app_name_image
assert push_snake_zrb_app_name_image



from typing import Optional

from module.auth.integration.model.permission_model import permission_model
from module.auth.schema.permission import PermissionData

DEFAULT_ACTION_NAMES = ["get", "get_by_id", "insert", "update", "delete"]
ACTION_DESCRIPTION = {
    "get": "list all",
    "get_by_id": "get single",
    "insert": "insert new",
    "update": "update",
    "delete": "delete",
}


async def ensure_entity_permission(
    module_name: str, entity_name: str, action_names: Optional[list[str]] = None
):
    if action_names is None:
        action_names = DEFAULT_ACTION_NAMES
    for action_name in action_names:
        caption = ACTION_DESCRIPTION[action_name]
        await permission_model.ensure_permission(
            PermissionData(
                name=f"{module_name}:{entity_name}:{action_name}",
                description=f"Allow {caption} {entity_name} on {module_name}",
            )
        )


async def register_permission():
    await ensure_entity_permission(module_name="auth", entity_name="group")
    await ensure_entity_permission(module_name="auth", entity_name="permission")
    await ensure_entity_permission(module_name="auth", entity_name="user")
    await ensure_entity_permission(
        module_name="log", entity_name="activity", action_names=["get", "get_by_id"]
    )



from zrb.helper.accessories.color import colored
from zrb.helper.log import logger
from zrb.helper.typecheck import typechecked
from zrb.task.task import Task

logger.debug(colored("Loading zrb.helper.task", attrs=["dark"]))


@typechecked
def show_lines(task: Task, *lines: str):
    separator = "\n    "
    task.print_out("\n" + separator + separator.join(lines) + "\n", trim_message=False)



from zrb.builtin._helper.reccuring_action import create_recurring_action
from zrb.runner import runner
from zrb.task.path_watcher import PathWatcher
from zrb.task.server import Controller, Server
from zrb.task_input.str_input import StrInput

watch_changes = Server(
    name="watch-changes",
    icon="🕵️",
    color="yellow",
    description="Watch changes and show message/run command",
    inputs=[
        StrInput(
            name="pattern",
            default="*.*",
            prompt="File pattern",
            description="File pattern to be watched",
        ),
        StrInput(
            name="ignored-pattern",
            default="",
            prompt="Ignored file pattern",
            description="Ignored file pattern",
        ),
    ],
    controllers=[
        Controller(
            name="watch",
            trigger=PathWatcher(
                name="watch-path",
                color="cyan",
                icon="👀",
                path="{{input.pattern}}",
                ignored_path="{{input.ignored_pattern}}",
            ),
            action=create_recurring_action(
                notif_title="Watch",
            ),
        )
    ],
)
runner.register(watch_changes)



from zrb.builtin.monorepo._config import MONOREPO_CONFIG
from zrb.builtin.monorepo._group import monorepo_group
from zrb.builtin.monorepo._helper import (
    create_add_subrepo_task,
    create_pull_monorepo_task,
    create_pull_subrepo_task,
    create_push_monorepo_task,
    create_push_subrepo_task,
)
from zrb.helper.util import to_kebab_case
from zrb.runner import runner
from zrb.task.flow_task import FlowTask
from zrb.task_group.group import Group

_pull_monorepo = create_pull_monorepo_task()

PULL_SUBREPO_UPSTREAM = _pull_monorepo
PUSH_SUBREPO_UPSTREAM = _pull_monorepo
for name, config in MONOREPO_CONFIG.items():
    kebab_name = to_kebab_case(name)
    group = Group(
        name=kebab_name, parent=monorepo_group, description=f"Subrepo {name} management"
    )
    origin = config.get("origin", "")
    folder = config.get("folder", "")
    branch = config.get("branch", "main")

    pull_subrepo = FlowTask(
        name="pull",
        group=group,
        upstreams=[_pull_monorepo],
        steps=[
            create_add_subrepo_task(
                origin=origin,
                folder=folder,
                branch=branch,
                task_name=f"add-{kebab_name}-subrepo",
            ),
            create_pull_subrepo_task(
                origin=origin,
                folder=folder,
                branch=branch,
                task_name=f"pull-{kebab_name}-subrepo",
            ),
        ],
    )
    runner.register(pull_subrepo)

    linked_pull_subrepo = FlowTask(
        name=f"pull-{kebab_name}",
        upstreams=[PULL_SUBREPO_UPSTREAM],
        steps=[
            create_add_subrepo_task(
                origin=origin,
                folder=folder,
                branch=branch,
                task_name=f"add-{kebab_name}-subrepo",
            ),
            create_pull_subrepo_task(
                origin=origin,
                folder=folder,
                branch=branch,
                task_name=f"pull-{kebab_name}-subrepo",
            ),
        ],
    )
    PULL_SUBREPO_UPSTREAM = linked_pull_subrepo

    push_subrepo = FlowTask(
        name="push",
        group=group,
        upstreams=[_pull_monorepo],
        steps=[
            create_add_subrepo_task(
                origin=origin,
                folder=folder,
                branch=branch,
                task_name=f"add-{kebab_name}-subrepo",
            ),
            create_pull_subrepo_task(
                origin=origin,
                folder=folder,
                branch=branch,
                task_name=f"pull-{kebab_name}-subrepo",
            ),
            create_push_subrepo_task(
                origin=origin,
                folder=folder,
                branch=branch,
                task_name=f"push-{kebab_name}-subrepo",
            ),
            create_push_monorepo_task(task_name="push-monorepo"),
        ],
    )
    runner.register(push_subrepo)

    linked_push_subrepo = FlowTask(
        name=f"push-{kebab_name}",
        upstreams=[PUSH_SUBREPO_UPSTREAM],
        steps=[
            create_add_subrepo_task(
                origin=origin,
                folder=folder,
                branch=branch,
                task_name=f"add-{kebab_name}-subrepo",
            ),
            create_pull_subrepo_task(
                origin=origin,
                folder=folder,
                branch=branch,
                task_name=f"pull-{kebab_name}-subrepo",
            ),
            create_push_subrepo_task(
                origin=origin,
                folder=folder,
                branch=branch,
                task_name=f"push-{kebab_name}-subrepo",
            ),
        ],
    )
    PUSH_SUBREPO_UPSTREAM = linked_push_subrepo


_push_monorepo = create_push_monorepo_task(task_name="push-monorepo")
PUSH_SUBREPO_UPSTREAM >> _push_monorepo
PUSH_SUBREPO_UPSTREAM = _push_monorepo



import socket
from collections.abc import Callable, Iterable
from typing import Any, Optional, TypeVar, Union

from zrb.helper.accessories.color import colored
from zrb.helper.log import logger
from zrb.helper.typecheck import typechecked
from zrb.helper.typing import JinjaTemplate
from zrb.task.any_task import AnyTask
from zrb.task.any_task_event_handler import (
    OnFailed,
    OnReady,
    OnRetry,
    OnSkipped,
    OnStarted,
    OnTriggered,
    OnWaiting,
)
from zrb.task.checker import Checker
from zrb.task_env.env import Env
from zrb.task_env.env_file import EnvFile
from zrb.task_group.group import Group
from zrb.task_input.any_input import AnyInput

logger.debug(colored("Loading zrb.task.port_checker", attrs=["dark"]))

TPortChecker = TypeVar("TPortChecker", bound="PortChecker")


@typechecked
class PortConfig:
    def __init__(self, host: str, port: int, timeout: int):
        self.host = host
        self.port = port
        self.timeout = timeout
        self.label = f"Checking {host}:{port}"


@typechecked
class PortChecker(Checker):
    def __init__(
        self,
        name: str = "check-port",
        group: Optional[Group] = None,
        inputs: Iterable[AnyInput] = [],
        envs: Iterable[Env] = [],
        env_files: Iterable[EnvFile] = [],
        icon: Optional[str] = None,
        color: Optional[str] = None,
        description: str = "",
        host: JinjaTemplate = "localhost",
        port: Union[int, JinjaTemplate] = 80,
        timeout: Union[int, JinjaTemplate] = 5,
        upstreams: Iterable[AnyTask] = [],
        fallbacks: Iterable[AnyTask] = [],
        on_triggered: Optional[OnTriggered] = None,
        on_waiting: Optional[OnWaiting] = None,
        on_skipped: Optional[OnSkipped] = None,
        on_started: Optional[OnStarted] = None,
        on_ready: Optional[OnReady] = None,
        on_retry: Optional[OnRetry] = None,
        on_failed: Optional[OnFailed] = None,
        checking_interval: Union[int, float] = 0.05,
        progress_interval: Union[int, float] = 5,
        expected_result: bool = True,
        should_execute: Union[bool, str, Callable[..., bool]] = True,
    ):
        Checker.__init__(
            self,
            name=name,
            group=group,
            inputs=inputs,
            envs=envs,
            env_files=env_files,
            icon=icon,
            color=color,
            description=description,
            upstreams=upstreams,
            fallbacks=fallbacks,
            on_triggered=on_triggered,
            on_waiting=on_waiting,
            on_skipped=on_skipped,
            on_started=on_started,
            on_ready=on_ready,
            on_retry=on_retry,
            on_failed=on_failed,
            checking_interval=checking_interval,
            progress_interval=progress_interval,
            expected_result=expected_result,
            should_execute=should_execute,
        )
        self._host = host
        self._port = port
        self._timeout = timeout
        self._config: Optional[PortChecker] = None

    def copy(self) -> TPortChecker:
        return super().copy()

    def to_function(
        self,
        env_prefix: str = "",
        raise_error: bool = True,
        is_async: bool = False,
        show_done_info: bool = True,
        should_clear_xcom: bool = False,
        should_stop_looper: bool = False,
    ) -> Callable[..., bool]:
        return super().to_function(
            env_prefix=env_prefix,
            raise_error=raise_error,
            is_async=is_async,
            show_done_info=show_done_info,
            should_clear_xcom=should_clear_xcom,
            should_stop_looper=should_stop_looper,
        )

    async def run(self, *args: Any, **kwargs: Any) -> bool:
        self._config = PortConfig(
            host=self.render_str(self._host),
            port=self.render_int(self._port),
            timeout=self.render_int(self._timeout),
        )
        return await super().run(*args, **kwargs)

    async def inspect(self, *args: Any, **kwargs: Any) -> bool:
        try:
            with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
                sock.settimeout(self._config.timeout)
                result = sock.connect_ex((self._config.host, self._config.port))
                if result == 0:
                    self.print_out(f"{self._config.label} (OK)")
                    return True
                self.show_progress(f"{self._config.label} (Not OK)")
        except Exception:
            self.show_progress(f"{self._config.label} (Socker error)")
        return False



import asyncio
import os
from typing import Any

from zrb.builtin.project._helper import validate_existing_project_dir
from zrb.builtin.project._input import project_dir_input
from zrb.builtin.project.add.fastapp._group import project_add_fastapp_group
from zrb.builtin.project.add.fastapp.app._input import app_name_input
from zrb.builtin.project.add.fastapp.crud._helper import (
    register_api,
    register_permission,
    register_rpc,
)
from zrb.builtin.project.add.fastapp.crud._input import (
    entity_name_input,
    main_column_name_input,
    plural_entity_name_input,
)
from zrb.builtin.project.add.fastapp.crud._task_factory import (
    create_add_navigation_task,
)
from zrb.builtin.project.add.fastapp.module._input import module_name_input
from zrb.helper import util
from zrb.helper.accessories.color import colored
from zrb.runner import runner
from zrb.task.cmd_task import CmdTask
from zrb.task.decorator import python_task
from zrb.task.resource_maker import ResourceMaker
from zrb.task.task import Task

_CURRENT_DIR = os.path.dirname(__file__)
_CODEMOD_DIR = os.path.join(_CURRENT_DIR, "nodejs", "codemod")


@python_task(
    name="validate",
    inputs=[project_dir_input, app_name_input, module_name_input, entity_name_input],
    retry=0,
)
async def validate(*args: Any, **kwargs: Any):
    project_dir = kwargs.get("project_dir")
    validate_existing_project_dir(project_dir)
    app_name = kwargs.get("app_name")
    module_name = kwargs.get("module_name")
    entity_name = kwargs.get("entity_name")
    snake_module_name = util.to_snake_case(module_name)
    snake_entity_name = util.to_snake_case(entity_name)
    automation_dir = os.path.join(
        project_dir, "_automate", util.to_snake_case(app_name)
    )
    if not os.path.exists(automation_dir):
        raise Exception(f"Automation directory does not exist: {automation_dir}")
    app_dir = os.path.join(project_dir, "src", f"{util.to_kebab_case(app_name)}")
    if not os.path.exists(app_dir):
        raise Exception(f"App directory does not exist: {app_dir}")
    module_path = os.path.join(app_dir, "src", "module", snake_module_name)
    if not os.path.exists(module_path):
        raise Exception(f"Module directory does not exist: {module_path}")
    entity_path = os.path.join(module_path, "entity", snake_entity_name)
    if os.path.exists(entity_path):
        raise Exception(f"Entity directory already exists: {entity_path}")


copy_resource = ResourceMaker(
    name="copy-resource",
    inputs=[
        project_dir_input,
        app_name_input,
        module_name_input,
        entity_name_input,
        plural_entity_name_input,
        main_column_name_input,
    ],
    upstreams=[validate],
    replacements={
        "zrbAppName": "{{input.app_name}}",
        "zrbModuleName": "{{input.module_name}}",
        "zrbEntityName": "{{input.entity_name}}",
        "zrbPluralEntityName": "{{input.plural_entity_name}}",
        "zrbColumnName": "{{input.column_name}}",
    },
    template_path=os.path.join(_CURRENT_DIR, "template"),
    destination_path="{{ input.project_dir }}",
    excludes=[
        "*/__pycache__",
    ],
)


@python_task(
    name="register-crud",
    inputs=[
        project_dir_input,
        app_name_input,
        module_name_input,
        entity_name_input,
        plural_entity_name_input,
        main_column_name_input,
    ],
    upstreams=[
        copy_resource,
    ],
)
async def register_crud(*args: Any, **kwargs: Any):
    task: Task = kwargs.get("_task")
    project_dir = kwargs.get("project_dir", ".")
    app_name = kwargs.get("app_name")
    module_name = kwargs.get("module_name")
    entity_name = kwargs.get("entity_name")
    await asyncio.gather(
        asyncio.create_task(
            register_api(task, project_dir, app_name, module_name, entity_name)
        ),
        asyncio.create_task(
            register_rpc(task, project_dir, app_name, module_name, entity_name)
        ),
        asyncio.create_task(
            register_permission(task, project_dir, app_name, module_name, entity_name)
        ),
    )


prepare_codemod = CmdTask(
    name="prepare-codemod",
    cwd=_CODEMOD_DIR,
    cmd=["npm install --save-dev", "node_modules/.bin/tsc"],
)

add_navigation = create_add_navigation_task(
    upstreams=[
        copy_resource,
        prepare_codemod,
    ]
)


@python_task(
    name="crud",
    group=project_add_fastapp_group,
    description="Add Fastapp CRUD",
    inputs=[
        project_dir_input,
        app_name_input,
        module_name_input,
        entity_name_input,
        plural_entity_name_input,
        main_column_name_input,
    ],
    upstreams=[register_crud, add_navigation],
    runner=runner,
)
async def add_fastapp_crud(*args: Any, **kwargs: Any):
    task: Task = kwargs.get("_task")
    task.print_out(colored("Fastapp crud added", color="yellow"))



from datetime import datetime

from zrb.builtin.monorepo._group import monorepo_group
from zrb.builtin.monorepo._task import PULL_SUBREPO_UPSTREAM
from zrb.runner import runner
from zrb.task.task import Task
from zrb.task_input.str_input import StrInput

pull_from_monorepo = Task(
    name="pull",
    group=monorepo_group,
    description="Pulling from subrepos",
    inputs=[
        StrInput(
            name="message",
            shortcut="m",
            prompt="Commit Messsage",
            default=lambda m: f"Synchronize subrepos at {datetime.now()}",
        )
    ],
    retry=0,
)
PULL_SUBREPO_UPSTREAM >> pull_from_monorepo
runner.register(pull_from_monorepo)



import os

from zrb.builtin.devtool.install._group import devtool_install_group
from zrb.builtin.devtool.install._helper import write_config
from zrb.runner import runner
from zrb.task.cmd_task import CmdTask
from zrb.task.flow_task import FlowTask
from zrb.task.task import Task

_CURRENT_DIR = os.path.dirname(__file__)
_SHELL_SCRIPT_DIR = os.path.join(_CURRENT_DIR, "..", "..", "..", "..", "shell-scripts")

install_helix = FlowTask(
    name="helix",
    group=devtool_install_group,
    description="Post modern text editor",
    steps=[
        CmdTask(
            name="install-helix",
            cmd_path=[
                os.path.join(_SHELL_SCRIPT_DIR, "_common-util.sh"),
                os.path.join(_CURRENT_DIR, "install.sh"),
            ],
            preexec_fn=None,
        ),
        [
            Task(
                name="create-helix-theme",
                run=write_config(
                    template_file=os.path.join(
                        _CURRENT_DIR,
                        "resource",
                        "themes",
                        "gruvbox_transparent.toml",  # noqa
                    ),
                    config_file="~/.config/helix/themes/gruvbox_transparent.toml",  # noqa
                    remove_old_config=True,
                ),
            ),
            Task(
                name="configure-helix",
                run=write_config(
                    template_file=os.path.join(_CURRENT_DIR, "resource", "config.toml"),
                    config_file="~/.config/helix/config.toml",
                    remove_old_config=True,
                ),
            ),
            CmdTask(
                name="install-language-server",
                cmd_path=[
                    os.path.join(_SHELL_SCRIPT_DIR, "_common-util.sh"),
                    os.path.join(_CURRENT_DIR, "install-language-server.sh"),
                ],
                preexec_fn=None,
            ),
        ],
    ],
    retry=0,
)
runner.register(install_helix)
